

#include "bf.h"

int bruteforce_pruning( point_set_t* ps_a, point_set_t* ps_b, point_set_t* ps_q)
{
	int i, j, k;
	point_t* p_a, *p_b, *p_q;
	DIST_TYPE dist_qa, dist_qb;

	for( i=0; i<ps_q->point_n; i++)
	{
		p_q = ps_q->point_array[ i];

		for( j=0; j<ps_a->point_n; j++)
		{
			p_a = ps_a->point_array[ j];

			dist_qa = calc_dist( p_q, p_a);

			for( k=0; k<ps_b->point_n; k++)
			{
				p_b = ps_b->point_array[ k];

				dist_qb = calc_dist( p_q, p_b);

				//
				if( dist_qb <= dist_qa)
					return 0;
			}//
		}
	}

	return 1;
}


rect_t* calc_best_fitting_rect( point_set_t* point_set_v)
{
	int i, j, dim;
	rect_t* rect_v;

	if( point_set_v->point_n == 0)
	{
		fprintf( stderr, "Empty point set [calc_best_fitting_rect].\n");
		exit( 0);
	}

	dim = point_set_v->point_array[ 0]->dim;

	rect_v = alloc_rect( dim);	

	for( i=0; i<dim; i++)
	{
		rect_v->axis[ i].max = -FLT_MAX;
		rect_v->axis[ i].min = FLT_MAX;

		for( j=0; j<point_set_v->point_n; j++)
		{
			if( point_set_v->point_array[ j]->coord[ i] > rect_v->axis[ i].max)
				rect_v->axis[ i].max = point_set_v->point_array[ j]->coord[ i];

			if( point_set_v->point_array[ j]->coord[ i] < rect_v->axis[ i].min)
				rect_v->axis[ i].min = point_set_v->point_array[ j]->coord[ i];
		}
	}
	
	return rect_v;
}

sphere_t* calc_best_fitting_sphere( point_set_t* point_set_v)
{
	//A simple O(n^2 d) algorithm is adopted here.
	
	int i, j, dim;
	sphere_t* sphere_v;
	point_t* point_v1, *point_v2, *point_v3, *point_v4;
	DIST_TYPE dist_max, dist_v;

	//
	if( point_set_v->point_n == 0)
	{
		fprintf( stderr, "Empty point set [calc_best_fitting_sphere].\n");
		exit( 0);
	}

	dim = point_set_v->point_array[ 0]->dim;

	sphere_v = alloc_sphere( dim);

	dist_max = 0;
	point_v3 = point_set_v->point_array[ 0];
	point_v4 = point_set_v->point_array[ 0];

	for( i=0; i<point_set_v->point_n-1; i++)
	{
		point_v1 = point_set_v->point_array[ i];

		for( j=i+1; j<point_set_v->point_n; j++)
		{
			point_v2 = point_set_v->point_array[ j];

			dist_v = calc_dist( point_v1, point_v2);

			if( dist_v > dist_max)
			{
				dist_max = dist_v;
				point_v3 = point_v1;
				point_v4 = point_v2;
			}
		}
	}
	
	//
	//sphere_v->radius = dist_max / 2;
	sphere_v->center = alloc_point( dim);
	for( i=0; i<dim; i++)
		sphere_v->center->coord[ i] = ( point_v3->coord[ i] + point_v4->coord[ i]) / 2;

	sphere_v->radius = 0;
	for( i=0; i<point_set_v->point_n; i++)
	{
		dist_v = calc_dist( sphere_v->center, point_set_v->point_array[ i]);

		if( dist_v > sphere_v->radius)
			sphere_v->radius = dist_v;
	}

	return sphere_v;
}