

#ifndef DATA_STRUCT_H
#define DATA_STRUCT_H

#include<math.h>
#include<float.h>
#include<string.h>
#include<stdio.h>
#include<stdlib.h>


#define COORD_TYPE	float
#define DIST_TYPE	double

#define	PRECISION_THR		0.000001
#define	DIST_PRECISION_THR	0.000001
#define PI					3.1415926535897932


#define abs(a, b)	( ( a) >= ( b) ? ( ( a)-b) : ( ( b) - ( a)))


//The data structure for storing points.
typedef struct point
{
	int			id;

	int			dim;
	COORD_TYPE*	coord;

}	point_t;

//The data structure for storing hypersheres.
typedef struct sphere
{
	int			id;

	point_t*	center;
	DIST_TYPE	radius;

}	sphere_t;

//The data structure for storing a range.
typedef struct range
{
	COORD_TYPE	min;
	COORD_TYPE	max;

}	range_t;

//The data structure for storing the data.
typedef struct sphere_data
{
	int			sphere_n;
	int			dim;

	sphere_t**	sphere_v;

	//For generating point_sets.
	range_t*	bounds;

}	sphere_data_t;

//The data structure for storing a rectangle.
typedef struct rect
{
	int			dim;
	range_t*	axis;

}	rect_t;

//The data structure for storing a set of points.
typedef struct point_set
{
	int			point_n;
	point_t**	point_array;

}	point_set_t;



//Added on 5 Dec., 2013
//For storing the "dominating set" and "dominated-by set", both correspond to sets of spheres
//Maintaining the pointers to the spheres is enough;
typedef struct sphere_node
{
	sphere_t*			sphere_v;
	
	struct sphere_node*	next;
}	sphere_node_t;

typedef struct sphere_list
{
	int				num;
	
	sphere_node_t*	head;
	sphere_node_t*	rear;

}	sphere_list_t;
	

#endif