

#include "data_utility.h"


#ifndef WIN32
/*
* get_curtime gets the current rusage info.
*
* @cur_t is used for storing the rusage information.
*/
void get_cur_time( rusage* cur_t)
{
	if( getrusage( RUSAGE_SELF, cur_t) != 0)
	{
		fprintf( stderr, "The invokation of getrusage fails.\n");
		exit( 0);
	}
}

/*
* GetTime calculates the time information.
* 
* @sta_t and end_t indicate the rusage information at the start and end, respectively.
* @user_t and system_t are used for storing the time information.
*
*/
void collect_time( struct rusage* sta_t, struct rusage* end_t, float* user_t)	//, float* system_t)
{
	(*user_t) = ((float)(end_t->ru_utime.tv_sec - sta_t->ru_utime.tv_sec)) + 
		((float)(end_t->ru_utime.tv_usec - sta_t->ru_utime.tv_usec)) * 1e-6;
/*
	(*system_t) = ((float)(end_t->ru_stime.tv_sec - sta_t->ru_stime.tv_sec)) +
		((float)(end_t->ru_stime.tv_usec - sta_t->ru_stime.tv_usec)) * 1e-6;
*/
}

#endif


/*
 *	Generates a number between [min, max] randomly.
 */
int rand_i( int min, int max)
{
	int i_rand;
	float ratio;

	if( min > max)
		return 0;

	if( max - min > RAND_MAX)
	{
		//printf( "rand_i: [%i, %i] out of range.\n", min, max);
		//exit( 0);

		ratio = rand_f( 0, 1);
		i_rand = min + ( int)( ratio * ( max - min));

		return i_rand;
	}

	//srand( time( NULL));

	if( min == max)
		i_rand = 0;
	else
		i_rand = rand( ) % ( max - min + 1);

	return i_rand + min;
}

/*
 *	Generate a random float value in [min_v, max_v].
 */
float rand_f( float min_v, float max_v)
{
	float rand_v;

	if( min_v > max_v)
		return 0;

	//srand( time(NULL));

	rand_v = float( rand( )) / RAND_MAX; //rand_v falls in [0,1] randomly.

	return ( max_v - min_v) * rand_v + min_v;
}


config_t* read_config( )
{
	config_t* config_v;
	FILE* c_fp;
	
	if( ( c_fp = fopen( CONFIG_FILE, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open the config file.\n");
		exit( 0);
	}

	config_v = ( config_t*)malloc( sizeof( config_t));
	memset( config_v, 0, sizeof( config_t));

	fscanf( c_fp, "%s%i%i", config_v->f_name, &config_v->dim, &config_v->sphere_n);
	//fscanf( c_fp, "%lf", &config_v->radius);
	//fscanf( c_fp, "%lf%lf", &config_v->mean_ratio, &config_v->var_ratio);
	//fscanf( c_fp, "%i%i", &config_v->query_n, &config_v->repeat_n);

	//For AkNN;
	//fscanf( c_fp, "%i%f%i", &config_v->k, &config_v->eps, &config_v->query_point_n);//, &config_v->dom_tag);

	//
	//fscanf( c_fp, "%i%i", &config_v->grid_scale, &config_v->point_n);

	fscanf( c_fp, "%i", &config_v->dom_tag);
	
	return config_v;
}

point_t* alloc_point( int dim)
{
	point_t* point_v;

	point_v = ( point_t*)malloc( sizeof( point_t));
	memset( point_v, 0, sizeof( point_t));

	point_v->dim = dim;

	point_v->coord = ( COORD_TYPE*)malloc( dim * sizeof( COORD_TYPE));
	memset( point_v->coord, 0, dim * sizeof( COORD_TYPE));

	return point_v;
}

void release_point( point_t* point_v)
{
	free( point_v->coord);
	free( point_v);
}

sphere_t* alloc_sphere( int dim)
{
	sphere_t* sphere_v;

	sphere_v = ( sphere_t*)malloc( sizeof( sphere_t));
	memset( sphere_v, 0, sizeof( sphere_t));

	sphere_v->center = alloc_point( dim);

	return sphere_v;
}

void ini_sphere( sphere_t* sphere_v)
{
	memset( sphere_v->center->coord, 0, sphere_v->center->dim * sizeof( COORD_TYPE));
	sphere_v->radius = 0;

	return;
}

void release_sphere( sphere_t* sphere_v)
{
	release_point( sphere_v->center);

	free( sphere_v);

	return;
}


sphere_data_t* alloc_sphere_data( int dim, int sphere_n)
{
	int i;

	sphere_data_t* sphere_data_v;

	sphere_data_v = ( sphere_data_t*)malloc( sizeof( sphere_data_t));
	memset( sphere_data_v, 0, sizeof( sphere_data_t));

	sphere_data_v->dim = dim;
	sphere_data_v->sphere_n = sphere_n;
	sphere_data_v->sphere_v = ( sphere_t**)malloc( sphere_n * sizeof( sphere_t*));
	memset( sphere_data_v->sphere_v, 0, sphere_n * sizeof( sphere_t*));

	for( i=0; i<sphere_n; i++)
	{
		sphere_data_v->sphere_v[ i] = alloc_sphere( dim);
	}

	//
	sphere_data_v->bounds = ( range_t*)malloc( dim * sizeof( range_t));
	memset( sphere_data_v->bounds, 0, dim * sizeof( range_t));

	return sphere_data_v;
}

sphere_data_t* read_sphere_data( config_t* config_v)
{
	int i, j;
	sphere_data_t* sphere_data_v;
	FILE* i_fp;

	if( ( i_fp = fopen( config_v->f_name, "r")) == NULL)
	{
		fprintf( stderr, "Cannot open %s [read_sphere_data].\n", config_v->f_name);
		exit( 0);
	}

	//Initialize
	sphere_data_v = alloc_sphere_data( config_v->dim, config_v->sphere_n);

	for( i=0; i<config_v->dim; i++)
	{
		sphere_data_v->bounds[ i].min = FLT_MAX;
		sphere_data_v->bounds[ i].max = FLT_MIN;
	}

	//Read
	for( i=0; i<config_v->sphere_n; i++)
	{
		fscanf( i_fp, "%i", &sphere_data_v->sphere_v[ i]->id);
		for( j=0; j<config_v->dim; j++)
		{
			fscanf( i_fp, "%f", &sphere_data_v->sphere_v[ i]->center->coord[ j]);

			//
			if( sphere_data_v->sphere_v[ i]->center->coord[ j] < sphere_data_v->bounds[ j].min)
				sphere_data_v->bounds[ j].min = sphere_data_v->sphere_v[ i]->center->coord[ j];
			if( sphere_data_v->sphere_v[ i]->center->coord[ j] > sphere_data_v->bounds[ j].max)
				sphere_data_v->bounds[ j].max = sphere_data_v->sphere_v[ i]->center->coord[ j];
		}

		fscanf( i_fp, "%lf", &sphere_data_v->sphere_v[ i]->radius);

		//testing.
		//sphere_data_v->sphere_v[ i]->radius += 200;
	}

	return sphere_data_v;
}

void release_sphere_data( sphere_data_t* sphere_data_v)
{
	int i;

	for( i=0; i<sphere_data_v->sphere_n; i++)
		release_sphere( sphere_data_v->sphere_v[ i]);

	free( sphere_data_v->sphere_v);
	free( sphere_data_v);
}

rect_t* alloc_rect( int dim)
{
	rect_t* rect_v;

	rect_v = ( rect_t*)malloc( sizeof( rect_t));
	memset( rect_v, 0, sizeof( rect_t));

	rect_v->dim = dim;
	rect_v->axis = ( range_t*)malloc( dim * sizeof( range_t));
	memset( rect_v->axis, 0, dim * sizeof( range_t));

	return rect_v;
}

void release_rect( rect_t* rect_v)
{
	free( rect_v->axis);
	free( rect_v);
}

point_set_t* alloc_point_set( int dim, int point_n)
{
	int i;
	point_set_t* point_set_v;

	point_set_v = ( point_set_t*)malloc( sizeof( point_set_t));
	memset( point_set_v, 0, sizeof( point_set_t));

	point_set_v->point_n = point_n;
	point_set_v->point_array = ( point_t**)malloc( point_n * sizeof( point_t*));
	memset( point_set_v->point_array, 0, point_n * sizeof( point_t*));

	for( i=0; i<point_n; i++)
	{
		point_set_v->point_array[ i] = alloc_point( dim);
	}

	return point_set_v;
}

void release_point_set( point_set_t* point_set_v)
{
	int i;

	for( i=0; i<point_set_v->point_n; i++)
	{
		release_point( point_set_v->point_array[ i]);
	}

	free( point_set_v->point_array);
	free( point_set_v);
}

void release_point_set_light( point_set_t* point_set_v)
{
	//int i;

	//for( i=0; i<point_set_v->point_n; i++)
	//{
	//	release_point( point_set_v->point_array[ i]);
	//}

	free( point_set_v->point_array);
	free( point_set_v);
}

DIST_TYPE calc_dist( point_t* point_v1, point_t* point_v2)
{
	int i, dim;
	DIST_TYPE diff;

	diff = 0;
	
	dim = point_v1->dim;
	for( i=0; i<dim; i++)
	{
		diff += ( DIST_TYPE)pow( point_v1->coord[ i] - point_v2->coord[ i], 2);
	}

	return ( DIST_TYPE)sqrt( diff);
}


DIST_TYPE calc_min_diff_lp( range_t* range_v, COORD_TYPE x, int p)
{
	COORD_TYPE diff;

	if( x < range_v->min)
		diff = range_v->min - x;
	else if( x <= range_v->max)
		diff = 0;
	else
		diff = range_v->max - x;
	
	return ( DIST_TYPE)pow( diff, p);
}

DIST_TYPE calc_max_diff_lp( range_t* range_v, COORD_TYPE x, int p)
{
	COORD_TYPE diff;

	if( x <= ( range_v->min + range_v->max) / 2)
		diff = range_v->max - x;
	else
		diff = x - range_v->min;
	
	return ( DIST_TYPE)pow( diff, p);
}

DIST_TYPE calc_min_dist( point_t* point_v, rect_t* rect_v)
{
	int i;
	DIST_TYPE dist;

	dist = 0;
	for( i=0; i<point_v->dim; i++)
	{
		dist += calc_min_diff_lp( &rect_v->axis[ i], point_v->coord[ i], 2);
	}

	return sqrt( dist);
}

int is_point_in_rect( point_t* point_v, rect_t* rect_v)
{
	int i;

	for( i=0; i<point_v->dim; i++)
	{
		if( point_v->coord[ i] < rect_v->axis[ i].min ||
			point_v->coord[ i] > rect_v->axis[ i].max)
			return 0;
	}

	return 1;
}

int is_sphere_in_rect( sphere_t* sphere_v, rect_t* rect_v)
{
	DIST_TYPE min_dist;

	min_dist = calc_min_dist( sphere_v->center, rect_v);

	if( min_dist >= sphere_v->radius)
		return 1;
	else
		return 0;
}

point_set_t* gen_point_set( sphere_data_t* sphere_data_v, int grid_scale, int point_n)
{
	int i, j, dim, rand_v, cnt;
	point_set_t* point_set_v;
	COORD_TYPE interval;

	rect_t* grid;
	int* point_array;

	dim = sphere_data_v->dim;

	//Randomly select a grid.
	grid = alloc_rect( dim);

	for( i=0; i<dim; i++)
	{
		rand_v = rand_i( 1, grid_scale);

		interval = ( sphere_data_v->bounds[ i].max - sphere_data_v->bounds[ i].min) / grid_scale;
		
		grid->axis[ i].min = sphere_data_v->bounds[ i].min + ( rand_v - 1) * interval;
		grid->axis[ i].max = grid->axis[ i].min + interval;
	}

	//Collect the data points within the grid.
	point_array = ( int*)malloc( sphere_data_v->sphere_n * sizeof( int));
	memset( point_array, 0, sphere_data_v->sphere_n * sizeof( int));
	cnt = 0;

	for( i=0; i<sphere_data_v->sphere_n; i++)
	{
		if( is_point_in_rect( sphere_data_v->sphere_v[ i]->center, grid))
			point_array[ cnt++] = i;
	}

	//Sample point_n points from the point_array.
	point_set_v = alloc_point_set( dim, point_n);
	for( i=0; i<point_n; i++)
	{
		rand_v = rand_i( 0, cnt-1);
		
		for( j=0; j<dim; j++)
			point_set_v->point_array[ i]->coord[ j] = sphere_data_v->sphere_v[ rand_v]->center->coord[ j];
	}

	//Release the resources.
	free( point_array);

	return point_set_v;
}

point_set_t* gen_point_set_v2( sphere_data_t* sphere_data_v, int point_n)
{
	int i, j, dim, rand_v, id_min;
	int* tag;
	
	point_t* point_v;
	DIST_TYPE dist_min, dist_v;
	point_set_t* point_set_v;

	dim = sphere_data_v->dim;
	point_set_v = alloc_point_set( dim, point_n);

	tag = ( int*)malloc( sphere_data_v->sphere_n * sizeof( int));
	memset( tag, 0, sphere_data_v->sphere_n * sizeof( int));

	//Randomly pick a point.
	rand_v = rand_i( 0, sphere_data_v->sphere_n-1);

	point_set_v->point_array[ 0]->id = sphere_data_v->sphere_v[ rand_v]->id;
	point_v = sphere_data_v->sphere_v[ rand_v]->center;
	for( j=0; j<dim; j++)
		point_set_v->point_array[ 0]->coord[ j] = point_v->coord[ j];
	tag[ rand_v] = 1;

	for( i=1; i<point_n; i++)
	{
		dist_min = DBL_MAX;
		for( j=0; j<sphere_data_v->sphere_n; j++)
		{
			if( tag[ j])
				continue;

			dist_v = calc_dist( point_v, sphere_data_v->sphere_v[ j]->center);
			if( dist_v < dist_min)
			{
				dist_min = dist_v;
				id_min = j;
			}
		}

		//
		point_set_v->point_array[ i]->id = sphere_data_v->sphere_v[ id_min]->id;
		for( j=0; j<dim; j++)
			point_set_v->point_array[ i]->coord[ j] = sphere_data_v->sphere_v[ id_min]->center->coord[ j];
		tag[ id_min] = 1;	
	}

	//Release the resources.
	free( tag);

	return point_set_v;
}

point_set_t* gen_point_set_v3( sphere_data_t* sphere_data_v, int point_n)
{
	int i, j, dim;
	point_set_t* point_set_v;

	dim = sphere_data_v->dim;
	point_set_v = alloc_point_set( dim, point_n);

	for( i=0; i<point_n; i++)
	{
		point_set_v->point_array[ i]->id = sphere_data_v->sphere_v[ i]->id;
		for( j=0; j<dim; j++)
			point_set_v->point_array[ i]->coord[ j] = sphere_data_v->sphere_v[ i]->center->coord[ j];
	}

	return point_set_v;
}



sphere_list_t* alloc_sphere_list( )
{
	sphere_list_t* sphere_list_v;

	sphere_list_v = ( sphere_list_t*)malloc( sizeof( sphere_list_t));
	memset( sphere_list_v, 0, sizeof( sphere_list_t));

	sphere_list_v->head = ( sphere_node_t*)malloc( sizeof( sphere_node_t));
	memset( sphere_list_v->head, 0, sizeof( sphere_node_t));

	sphere_list_v->rear = sphere_list_v->head;

	return sphere_list_v;
}

void add_entry_sphere_list( sphere_list_t* sphere_list_v, sphere_t* sphere_v)
{
	sphere_node_t* sphere_node_v;

	sphere_node_v = ( sphere_node_t*)malloc( sizeof( sphere_node_t));
	memset( sphere_node_v, 0, sizeof( sphere_node_t));

	sphere_node_v->sphere_v = sphere_v;

	sphere_list_v->rear->next = sphere_node_v;
	
	sphere_list_v->rear = sphere_node_v;
	sphere_list_v->num ++;
}

void release_sphere_list( sphere_list_t* sphere_list_v)
{
	sphere_node_t* iter, *iter_2;

	iter = sphere_list_v->head;
	while( iter != NULL)
	{
		iter_2 = iter->next;

		free( iter);
		
		iter = iter_2;
	}

	free( sphere_list_v);
	return;
}


DIST_TYPE calc_min_min_dist( sphere_t* sphere_v1, sphere_t* sphere_v2)
{
	DIST_TYPE dist_v;

	dist_v = calc_dist( sphere_v1->center, sphere_v2->center) - sphere_v1->radius - sphere_v2->radius;

	if( dist_v < 0)
		return 0;
	else 
		return dist_v;
}

void adjust_sphere_radius( sphere_data_t* sphere_data_v, double mean_ratio, double var_ratio)
{
	int i, j, dim;

	DIST_TYPE min, max, low, high, dist_max, mean, var;
	_range_t* MBR;

	dim = sphere_data_v->dim;

	//Compute the MBR of the spheres;
	MBR = ( _range_t*)malloc( dim * sizeof( _range_t));
	memset( MBR, 0, dim * sizeof( _range_t));

	for( i=0; i<dim; i++)
	{
		min = DBL_MAX;
		max = -DBL_MAX;
	
		for( j=0; j<sphere_data_v->sphere_n; j++)
		{
			low = sphere_data_v->sphere_v[ j]->center->coord[ i];
				//- sphere_data_v->sphere_v[ j]->radius;
			high = sphere_data_v->sphere_v[ j]->center->coord[ i];
				//+ sphere_data_v->sphere_v[ j]->radius;
			
			if( low < min)
				min = low;
			if( high > max)
				max = high;
		}

		MBR[ i].max = ( COORD_TYPE)max;
		MBR[ i].min = ( COORD_TYPE)min;
	}

	//Specify the "absolute values" of mean and var;
	dist_max = 0;
	for( i=0; i<dim; i++)
		dist_max += pow( MBR[ i].max - MBR[ i].min, 2);
	dist_max = sqrt( dist_max);

	mean = mean_ratio * dist_max;
	var = var_ratio * mean;

	//Adjust the radii;
	for( i=0; i<sphere_data_v->sphere_n; i++)
		sphere_data_v->sphere_v[ i]->radius = gaussian_f( ( float)mean, ( float)var);
	
	//
	free( MBR);
	return;
}

void set_sphere_radius_zero( sphere_data_t* sphere_data_v)
{
	int i;

	for( i=0; i<sphere_data_v->sphere_n; i++)
		sphere_data_v->sphere_v[ i]->radius = 0;
}

float gaussian_f( float mean, float s_var)
{
	float x1, x2, w, y1, y2;

	do
	{
		x1 = 2 * rand_f( 0, 1) - 1;
		x2 = 2 * rand_f( 0, 1) - 1;
		w = x1 * x1 + x2 * x2;
	} while( w >= 1);

	w = float(sqrt( float(-2.0 * log(w)) / w));
	y1 = x1 * w;					//y1 follows N[0, 1^2].

	y2 = y1 * s_var + mean;			//y2 follows N[mean, s_var^2].

	return y2;
}

void print_point( FILE* fp, point_t* point_v, int line_break_tag)
{
	fprintf( fp, "%i  ", point_v->id);

	if( line_break_tag == 1)
		fprintf( fp, "\n");
}


void print_point_set( FILE* fp, point_set_t* point_set_v, int line_break_tag)
{
	int i;

	for( i=0; i<point_set_v->point_n; i++)
		print_point( fp, point_set_v->point_array[ i], 0);

	fprintf( fp, "\n");

	if( line_break_tag == 1)
		fprintf( fp, "\n");
}

void print_sphere( FILE* fp, sphere_t* sphere_v, int line_break_tag)
{
	int i;

	fprintf( fp, "%i\t", sphere_v->id);
	for( i=0; i<sphere_v->center->dim; i++)
		fprintf( fp, "%0.2f  ", sphere_v->center->coord[ i]);
	fprintf( fp, "\t%0.2f", sphere_v->radius);
	fprintf( fp, "\n");
}
