

#ifndef DATA_UTILITY_H
#define DATA_UTILITY_H


#include "data_struct.h"
#include "sstree.h"

#define	MAX_FILENAME_LENG	256
#define CONFIG_FILE			"config.txt"
#define STAT_FILE			"stat.txt"

typedef struct config
{
	char	f_name[ MAX_FILENAME_LENG];
	int		sphere_n;
	int		dim;

	//DIST_TYPE	radius;
	double	mean_ratio;
	double	var_ratio;
	
	int		query_n;
	int		repeat_n;

	//For AkNN;
	int		k;
	float	eps;
	int		query_point_n;

	//int		dom_tag;

	//For querying point sets.
	//int		grid_scale;
	//int		point_n;

	int		dom_tag;

}	config_t;

typedef struct stat
{
	float	trigonometric_time;
	float	hyperbola_time;

	int		TP;
	int		FP;
	int		TN;
	int		FN;

	//Only for querying the point sets.
	float	bf_time;

	int		TP_2;
	int		FP_2;
	int		TN_2;
	int		FN_2;

	//Exp 1: For Spatial Domimance
	float	Dom_time[ 5];
	float	Dom_precision[ 5];
	float	Dom_recall[ 5];

	float	Dom_TP[ 5];
	float	Dom_TN[ 5];
	float	Dom_FP[ 5];
	float	Dom_FN[ 5];

	//Exp 2. For "Basic Dominance" computing
	float	DS_non_index_time[ 5];
	float	DS_index_time[ 5];
	float	DS_index_Dom_num_node[ 5];
	float	DS_index_Dom_num_obj[ 5];
	//float	DS_index_pa[ 5];
	//float	DS_index_hn_MBC[ 5];
	//float	DS_index_hn_point[ 5];

	float	DSB_non_index_time[ 5];
	float	DSB_index_time[ 5];
	float	DSB_index_Dom_num_node[ 5];
	float	DSB_index_Dom_num_obj[ 5];
	//float	DSB_index_pa[ 5];
	//float	DSB_index_hn_MBC[ 5];
	//float	DSB_index_hn_point[ 5];


	//Exp 3. For AkNN
	float	AkNN_time[ 5];

	float	AkNN_hn_MBC[ 5];
	float	AkNN_hn_point[ 5];
	float	AkNN_pa[ 5];

	//float	AkNN_pa_cnt_0;
	//float	AkNN_pa_cnt_1;

}	stat_t;
	

#ifndef WIN32
#include <sys/resource.h>

void get_cur_time( rusage* cur_t);

void collect_time( struct rusage* sta_t, struct rusage* end_t, float* user_t);	//, float* system_t);
#endif

int rand_i( int min, int max);

float rand_f( float min_v, float max_v);

config_t* read_config( );

point_t* alloc_point( int dim);

void release_point( point_t* point_v);

sphere_t* alloc_sphere( int dim);

void ini_sphere( sphere_t* sphere_v);

void release_sphere( sphere_t* sphere_v);

sphere_data_t* alloc_sphere_data( int dim, int sphere_n);

sphere_data_t* read_sphere_data( config_t* config_v);

void release_sphere_data( sphere_data_t* sphere_data_v);

rect_t* alloc_rect( int dim);

void release_rect( rect_t* rect_v);

point_set_t* alloc_point_set( int dim, int point_n);

void release_point_set( point_set_t* point_set_v);
void release_point_set_light( point_set_t* point_set_v);

DIST_TYPE calc_dist( point_t* point_v1, point_t* point_v2);

DIST_TYPE calc_min_diff_lp( range_t* range_v, COORD_TYPE x, int p);

DIST_TYPE calc_max_diff_lp( range_t* range_v, COORD_TYPE x, int p);

DIST_TYPE calc_min_dist( point_t* point_v, rect_t* rect_v);

int is_point_in_rect( point_t* point_v, rect_t* rect_v);

int is_sphere_in_rect( sphere_t* sphere_v, rect_t* rect_v);

/*
 *	Sample a set of point_n points from a random grid.
 */
point_set_t* gen_point_set( sphere_data_t* sphere_data_v, int grid_scale, int point_n);

/*
 *	Use a set of the point_n nearest points from a random point.
 */
point_set_t* gen_point_set_v2( sphere_data_t* sphere_data_v, int point_n);

point_set_t* gen_point_set_v3( sphere_data_t* sphere_data_v, int point_n);

//
sphere_list_t* alloc_sphere_list( );

void add_entry_sphere_list( sphere_list_t* sphere_list_v, sphere_t* sphere_v);

void release_sphere_list( sphere_list_t* sphere_list_v);


//Added on 6 Dec. 2013
DIST_TYPE calc_min_min_dist( sphere_t* sphere_v1, sphere_t* sphere_v2);

void adjust_sphere_radius( sphere_data_t* sphere_data_v, double mean_ratio, double var_ratio);

void set_sphere_radius_zero( sphere_data_t* sphere_data_v);

float gaussian_f( float mean, float s_var);


void print_point( FILE* fp, point_t* point_v, int line_break_tag);

void print_point_set( FILE* fp, point_set_t* point_set_v, int line_break_tag);

void print_sphere( FILE* fp, sphere_t* sphere_v, int line_break_tag);

#endif