

#include "hyperbola.h"


point_t* dup_point( point_t* point_v)
{
	int i;
	point_t* point_v2;

	point_v2 = alloc_point( point_v->dim);

	for( i=0; i<point_v->dim; i++)
		point_v2->coord[ i] = point_v->coord[ i];

	return point_v2;
}

int hyperbola_pruning( sphere_t* s_a, sphere_t* s_b, sphere_t* s_q)
{
	int i, j, dim, sol_num;
	DIST_TYPE d_min, dist_ab, dist_qb, dist_qa, alpha, r, dist_v;
	double a_1, a_2, a_3, a_4, a_5;
	double alpha_square, four_times_r_square, qa_square, qb_square;

	//double A, B, C, D, E;
	double coe[ 5], sol[ 4], soli[4], lambda;
	
	point_t *c_a, *c_b, *c_q, *x;

	if( is_spheres_overlap( s_a, s_b))
	{
		//s_a and s_b overlap.
		return 0;
	}
	else if( is_spheres_overlap( s_b, s_q))
	{
		return 0;
	}
	else
	{
		//The other case.
		//P <-- the hyperbola
		//Check whether the s_q falls in the region 
		//which is formed by P and contains c_a with two steps.

		dim = s_q->center->dim;

		//bug.
		/*
		c_a = s_a->center;
		c_b = s_b->center;
		c_q = s_q->center;
		*/

		c_a = dup_point( s_a->center);	//Could be optimized.
		c_b = dup_point( s_b->center);
		c_q = dup_point( s_q->center);		

		dist_ab = calc_dist( c_a, c_b);
		dist_qa = calc_dist( c_q, c_a);
		dist_qb = calc_dist( c_q, c_b);

		//Step 0: Check whether c_q is within the region.
		if( dist_qb - dist_qa <= s_a->radius + s_b->radius)
			return 0;

		//Step 1: Compute the minimum distance between P and c_q, denoted by d_min;
		alpha = dist_ab / 2;
		r = s_a->radius + s_b->radius;


		//emp
		qa_square = dist_qa * dist_qa;
		qb_square = dist_qb * dist_qb;
		alpha_square = alpha * alpha;
		four_times_r_square = 4 * r * r;

		//Coordinate shifting.
		//c_a --> (-alpha, 0, 0, ...)
		//c_b --> (alpha, 0, 0, ...)
		//c_q --> ( ( dist_qa^2 - dist_qb^2) / ( 4 * alpha), sqrt( dist_qa^2 - ( ( dist_qa^2 - dist_qb^2 + 4 * alpha^2) / ( 4 * alpha))^2), 0, 0, ...)

		memset( c_a->coord, 0, dim * sizeof( COORD_TYPE));
		c_a->coord[ 0] = - ( COORD_TYPE)alpha;

		memset( c_b->coord, 0, dim * sizeof( COORD_TYPE));
		c_b->coord[ 0] = ( COORD_TYPE)alpha;

		memset( c_q->coord, 0, dim * sizeof( COORD_TYPE));
		c_q->coord[ 0] = ( COORD_TYPE)( ( qa_square - qb_square) / ( 4 * alpha));
		//t
		//printf( "%lf\n", dist_qa * dist_qa - pow( c_q->coord[ 0] + alpha, 2));
		//t
		if( qa_square - pow( c_q->coord[ 0] + alpha, 2) < 0)
			c_q->coord[ 1] = 0;
		else
			c_q->coord[ 1] = ( COORD_TYPE)sqrt( qa_square - pow( c_q->coord[ 0] + alpha, 2));

		//bug.
		if( fabs( r) < 1e-12)
		{
			//It implies that both s_a and s_b are points;
			if( fabs( c_q->coord[ 0]) > alpha / 2)
				return 1;
			else
				return 0;
		}

		/*
		DIST_TYPE dist_ab_1, dist_qa_1, dist_qb_1;
		dist_ab_1 = calc_dist( c_a, c_b);
		dist_qa_1 = calc_dist( c_q, c_a);
		dist_qb_1 = calc_dist( c_q, c_b);
		*/

		//Maintain a_1, a_2, a_3, a_4, a_5;
		a_1 = ( 16 * alpha_square - four_times_r_square) * pow( c_q->coord[ 0], 2);
		a_2 = pow( r, 4) - four_times_r_square * alpha_square;
		a_3 = 0;
		for( i=1; i<dim; i++)
			a_3 += pow( c_q->coord[ i], 2);
		a_3 *= four_times_r_square;
		a_4 = four_times_r_square;
		a_5 = four_times_r_square - 16 * alpha_square;

		//Maintain A, B, C, D, E;
		coe[4] = a_2 * a_4 * a_4 * a_5 * a_5;
		coe[3] = 2 * a_2 * a_4 * a_4 * a_5 + 2 * a_2 * a_4 * a_5 * a_5;
		coe[2] = a_1 * a_4 * a_4 + a_2 * a_4 * a_4 + 4 * a_2 * a_4 * a_5 + a_2 * a_5 * a_5 - a_3 * a_5 * a_5;
		coe[1] = 2 * a_1 * a_4 + 2 * a_2 * a_4 + 2 * a_2 * a_5 - 2 * a_3 * a_5;
		coe[0] = a_1 + a_2 - a_3;

		/*//t
		FILE* t_fp;
		
		t_fp = fopen( "test_result.txt", "w");
		for( i=0; i<5; i++)
			fprintf( t_fp, "%f\n", coe[4-i]);
		fprintf( t_fp, "\n");
		//t*/


		//Solve the quartic equation with the coefficients of A, B, C, D, and E.
		sol_num = 0;
		//sol_num = solveQuartic( coe, sol);
		quartic( coe, sol, soli, &sol_num);

		/*//t
		for( i=0; i<sol_num; i++)
		{
			fprintf( t_fp, "%f\n", sol[ i]);
		}
		fclose( t_fp);
		//t*/

		x = alloc_point( dim);

		//Compute d_min;
		d_min = DBL_MAX;
		for( i=0; i<sol_num; i++)
		{
			lambda = sol[ i];

			//
			x->coord[ 0] = ( COORD_TYPE)( c_q->coord[ 0] / ( 1 + four_times_r_square * lambda - 16 * alpha_square * lambda));
			for( j=1; j<dim; j++)
			{
				x->coord[ j] = ( COORD_TYPE)( c_q->coord[ j] / ( four_times_r_square * lambda + 1));
			}

			//
			dist_v = calc_dist( c_q, x);

			if( dist_v < d_min)
				d_min = dist_v;
		}

		release_point( x);
		
		release_point( c_a);
		release_point( c_b);
		release_point( c_q);

		//Step 2: Perform the decision.
		if( d_min > s_q->radius)	// && ( dist_qb - dist_qa) > r
			return 1;
		else
			return 0;
	}
}
