

#include "kNN.h"


//Application 1: basic dominance count computing.
int comp_sphere_dom( sphere_t* s_a, sphere_t* s_b, sphere_t* s_q, int dom_opt)
{
	if( dom_opt == 0)
	{
		return hyperbola_pruning( s_a, s_b, s_q);
	}
	else if( dom_opt == 1)
	{
		return MaxMin_pruning( s_a, s_b, s_q);
	}
	else if( dom_opt == 2)
	{
		return MBR_pruning( s_a, s_b, s_q);
	}
	else if( dom_opt == 3)
	{
		return trigonometric_pruning( s_a, s_b, s_q);
	}
	else
	{
		fprintf( stderr, "Invalid dom_tag [comp_sphere_dom]\n");
		exit( 0);
	}
}

sphere_list_t* comp_basic_DS( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int alg_opt, int dom_opt)
{
	if( alg_opt == 1)
		return comp_basic_DS_linear_scan( sphere_data_v, s_a, s_q, prob_tag, dom_opt);
	else
		return comp_basic_DS_index( sphere_data_v, s_a, s_q, prob_tag, dom_opt);
}

sphere_list_t* comp_basic_DS_linear_scan( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int dom_opt)
{
	int i, res;
	sphere_t* sphere_v;
	sphere_list_t* basic_DS;

	//
	basic_DS = alloc_sphere_list( );

	//
	for( i=0; i<sphere_data_v->sphere_n; i++)
	{
		sphere_v = sphere_data_v->sphere_v[ i];

		if( prob_tag == 1)	//""
			res = comp_sphere_dom( sphere_v, s_a, s_q, dom_opt);
		else //
			res = comp_sphere_dom( s_a, sphere_v, s_q, dom_opt);

		if( res == 1)
		{
			add_entry_sphere_list( basic_DS, sphere_v);
		}
	}	
	
	return basic_DS;
}

sphere_list_t* comp_basic_DS_index( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int dom_opt)
{
	int i, size, top, res, cnt;
	b_heap_t* b_h;

	sphere_t* MBC, *MBC_cur;
	node_t* node_v;
	obj_t* obj_v;

	sphere_list_t* sphere_list_v;

	//
	sphere_list_v = alloc_sphere_list( );

	//Build a SS-Tree on the set of spheres;
	ini_tree_ss( );
	build_SSTree( sphere_data_v);

	//Perform the "best-first" search;
	size = sphere_data_v->sphere_n;
	
	calc_MBC_ss( SS_Tree_v.root, MBC, 0);

	//
	cnt = 1;
	b_h = alloc_b_heap( size);
	
	b_h->obj_arr[ cnt].node_v = SS_Tree_v.root;
	b_h->obj_arr[ cnt].key = calc_min_min_dist( MBC, s_q);

	insert_b_heap( b_h, cnt++);

	//
	while( b_h->rear > 0)
	{
		top = get_top( b_h);

		node_v = b_h->obj_arr[ top].node_v;

		//Check the "dominance";
		if( node_v->parent == NULL)
			MBC_cur = MBC;
		else
			MBC_cur = node_v->parent->MBC_array[ node_v->loc];

		if( prob_tag == 1)
			res = comp_sphere_dom( MBC_cur, s_a, s_q, dom_opt);
		else
			res = comp_sphere_dom( s_a, MBC_cur, s_q, dom_opt);

		//t
		if( prob_tag == 1)
			stat_v_DS.DS_index_Dom_num_node[ dom_opt] ++;
		else
			stat_v_DS.DSB_index_Dom_num_node[ dom_opt] ++;
		//t

		if( res == 1)
		{
			//Traverse the sub-tree;
			traverse_SSTree( node_v, sphere_list_v);
		}
		else
		{
			//
			if( node_v->level > 0)
			{
				//node_v is an inner node;
				//Equeue all children of node_v;
				for( i=0; i<node_v->num; i++)
				{
					b_h->obj_arr[ cnt].node_v = ( node_t*)node_v->child[ i];
					b_h->obj_arr[ cnt].key = calc_min_min_dist( node_v->MBC_array[ i], s_q);

					insert_b_heap( b_h, cnt++);
				}	
			}
			else
			{
				//node_v is a leaf node;
				//Further check all the objects under node_v;
				for( i=0; i<node_v->num; i++)
				{
					obj_v = ( obj_t*)node_v->child[ i];
					
					if( prob_tag == 1)
						res = comp_sphere_dom( obj_v->MBC, s_a, s_q, dom_opt);
					else
						res = comp_sphere_dom( s_a, obj_v->MBC, s_q, dom_opt); 

					if( res)
						add_entry_sphere_list( sphere_list_v, obj_v->MBC);

					//t
					if( prob_tag == 1)
						stat_v_DS.DS_index_Dom_num_obj[ dom_opt] ++;
					else
						stat_v_DS.DSB_index_Dom_num_obj[ dom_opt] ++;
					//t
				}
			}//else
		}//else
	}//while

	//Release the resources;
	free_SSTree( );
	release_b_heap( b_h);

	release_sphere( MBC);	

	return sphere_list_v;
}


void emp_basic_DB( )
{
	int p, a, d, i, f_tag, ins_n, rand_v1, rand_v2;
	config_t* config_v;
	sphere_data_t* S;
	
	//
	sphere_t* s_a, *s_q;

	FILE* s_fp;

	//
	memset( &stat_v_DS, 0, sizeof( stat_t));
	f_tag = 0;

	//Read the hyperspheres;
	config_v = read_config( );

	S = read_sphere_data( config_v);

	adjust_sphere_radius( S, config_v->mean_ratio, config_v->var_ratio);

	//Build the SS-tree;
	ini_tree( );
	build_SSTree( S);

	//Change all the hyperspheres to be points;
	//set_sphere_radius_zero( S);

	ins_n = 0;
	for( i=0; i<config_v->query_n; i++)
	{
		ins_n ++;
		//if( ins_n % 100 == 0)
		printf( "Instance %i...\n", ins_n);

		rand_v1 = rand_i( 1, S->sphere_n);
		//rand_v2 = rand_i( 1, S->sphere_n);
		rand_v2 = rand_i( 1, S->sphere_n);

		s_a = S->sphere_v[ rand_v1];
		//S_b = S->sphere_v[ rand_v2];
		s_q = S->sphere_v[ rand_v2];

		//Handling "Basic_DS" problem;
		for( p=1; p<=2; p++)
		{
			for( a=1; a<=2; a++)
			{
				for( d=0; d<5; d++)
				{
#ifndef WIN32
					float usr_t;
					struct rusage start_t, end_t;
					
					get_cur_time( &start_t);
#endif
					comp_basic_DS( S, s_a, s_q, p, a, d);
					
#ifndef WIN32
					get_cur_time( &end_t);
					collect_time( &start_t, &end_t, &usr_t);
					if( p == 1 && a == 1)
					{
						//DS && non-index;
						stat_v_DS.DS_non_index_time[ d] += usr_t;
					}
					else if( p == 1 && a == 2)
					{
						//DS && non-index;
						stat_v_DS.DS_index_time[ d] += usr_t;
					}
					else if( p == 2 && a == 1)
					{
						//DS && non-index;
						stat_v_DS.DSB_non_index_time[ d] += usr_t;
					}
					else
					{
						//DS && non-index;
						stat_v_DS.DSB_index_time[ d] += usr_t;
					}
#endif
				}//for( d)
			}//for( a)
		}//for( p)
	}//for( i)

	//Print the statistics;
	if( ( s_fp = fopen( STAT_FILE, "w")) == NULL)
	{
		fprintf( stderr, "Cannot open the stat.txt file.\n");
		exit( 0);
	}

	//t
	for( i=0; i<5; i++)
	{
		fprintf( s_fp, "%f\n%f\n%i\n%.0f\n%.0f\n\n%f\n%f\n%i\n%.0f\n%.0f\n\n", 
						stat_v_DS.DS_non_index_time[ i] / config_v->query_n, 
						stat_v_DS.DS_index_time[ i] / config_v->query_n,
						S->sphere_n-1, 
						stat_v_DS.DS_index_Dom_num_node[ i] / config_v->query_n,
						stat_v_DS.DS_index_Dom_num_obj[ i] / config_v->query_n,
						stat_v_DS.DSB_non_index_time[ i] / config_v->query_n,
						stat_v_DS.DSB_index_time[ i] / config_v->query_n,
						S->sphere_n-1, 
						stat_v_DS.DSB_index_Dom_num_node[ i] / config_v->query_n,
						stat_v_DS.DSB_index_Dom_num_obj[ i] / config_v->query_n);
	}
	
	printf( "Time (non-index)\tDom num (non-index)\tTime (index)\tDom num (index)\n");
	for( i=0; i<5; i++)
	{
		fprintf( stdout, "%f\t%f\t%i\t%.0f\t%.0f\t\t%f\t%f\t%i\t%.0f\t%.0f\n", 
						stat_v_DS.DS_non_index_time[ i] / config_v->query_n, 
						stat_v_DS.DS_index_time[ i] / config_v->query_n,
						S->sphere_n-1, 
						stat_v_DS.DS_index_Dom_num_node[ i] / config_v->query_n,
						stat_v_DS.DS_index_Dom_num_obj[ i] / config_v->query_n,
						stat_v_DS.DSB_non_index_time[ i] / config_v->query_n, 
						stat_v_DS.DSB_index_time[ i] / config_v->query_n,
						S->sphere_n-1, 
						stat_v_DS.DSB_index_Dom_num_node[ i] / config_v->query_n,
						stat_v_DS.DSB_index_Dom_num_obj[ i] / config_v->query_n);
	}

	//Release the resources;
	//release_point_set( R);
	release_sphere_data( S);
	free( config_v);
	fclose( s_fp);
}


//Application 2: AkNN;

sphere_t* const_sphere_point( point_t* point_v)
{
	sphere_t* sphere_v;

	//sphere_v = alloc_sphere( point_v->dim);

	sphere_v = ( sphere_t*)malloc( sizeof( sphere_t));
	memset( sphere_v, 0, sizeof( sphere_t));
	
	sphere_v->center = point_v;
	sphere_v->radius = 0;

	return sphere_v;
}

void destroy_sphere_point( sphere_t* sphere_v)
{
	free( sphere_v);

	return;
}

point_set_t** AkNN( sphere_data_t* S, point_set_t* R, float eps, int k, int dom_tag)
{
	int i, j, dim, cnt, query_point_n, top, res;
	int* cnt_2;

	DIST_TYPE dist_v;
	DIST_TYPE* dist_k;
	sphere_t* s_q, *MBC, *MBC_cur;

	node_t* node_c;

	point_set_t* pruning_list;
	b_heap_t* candidate_queue;
	b_heap_t** solution_queue;

	point_set_t** solution_set;

	//
	dim = S->dim;
	query_point_n = R->point_n;

	solution_set = ( point_set_t**)malloc( query_point_n * sizeof( point_set_t*));
	memset( solution_set, 0, query_point_n * sizeof( point_set_t*));
	
	//Build the SSTree;
	ini_tree( );
	build_SSTree( S);
	
	//Prepare the sphere for the set of query points;
	s_q = calc_best_fitting_sphere( R);

	//Initialize the data structures for the algorithm;
	if( eps < 1)
	{
		fprintf( stderr, "Invalid parameter eps [AkNN].\n");
		exit( 0);
	}

	//1. pruning_list;
	//pruning_list = alloc_point_set( dim, ( int)( eps * k));
	pruning_list = comp_pruning_list( eps, k, s_q);

	//t
	//print_point_set( stdout, pruning_list, 1);
	//t

	//2. candidate_queue;
	candidate_queue = alloc_b_heap( 2 * S->sphere_n + 1);

	calc_MBC_ss( SS_Tree_v.root, MBC, 0);
	
	cnt = 1;
	candidate_queue->obj_arr[ cnt].node_v = SS_Tree_v.root;
	candidate_queue->obj_arr[ cnt].key = calc_min_min_dist( MBC, s_q);

	insert_b_heap( candidate_queue, cnt++);

	//3. solution_queue;
	solution_queue = ( b_heap_t**)malloc( query_point_n * sizeof( b_heap_t*));
	memset( solution_queue, 0, query_point_n * sizeof( b_heap_t*));

	cnt_2 = ( int*)malloc( query_point_n * sizeof( int));
	memset( cnt_2, 0, query_point_n * sizeof( int));

	dist_k = ( DIST_TYPE*)malloc( query_point_n * sizeof( DIST_TYPE));
	memset( dist_k, 0, query_point_n * sizeof( DIST_TYPE));

	for( i=0; i<query_point_n; i++)
	{
		dist_k[ i] = DBL_MAX;

		//
		solution_queue[ i] = alloc_b_heap( k + 1);
			
		cnt_2[ i] = 1;

		/**/
		for( j=0; j<pruning_list->point_n; j++)
		{	
			dist_v = calc_dist( pruning_list->point_array[ j], R->point_array[ i]);

			//if( dist_v < dist_k[ i])
			//{
				//Initailize the solution_queue based on pruning_list;
				if( solution_queue[ i]->rear == k)
					cnt_2[ i] = get_top( solution_queue[ i]);
				
				//
				solution_queue[ i]->obj_arr[ cnt_2[ i]].point_v = pruning_list->point_array[ j];
				solution_queue[ i]->obj_arr[ cnt_2[ i]].key = - dist_v;
				
				insert_b_heap( solution_queue[ i], cnt_2[ i]);
				
				//
				if( solution_queue[ i]->rear < k)
					cnt_2[ i] ++;

				//
				dist_k[ i] = - solution_queue[ i]->obj_arr[ solution_queue[ i]->h_arr[ 1]].key;
			//}//if
		}//for
		/**/
	}

	//Perform the "best-first" search;
	while( candidate_queue->rear > 0)
	{
		top = get_top( candidate_queue);

		//t
		//Collect the page access statistics;
		stat_v_AkNN.AkNN_pa[ dom_tag] ++;
		//t

		//
		for( i=0; i<query_point_n; i++)
		{
			if( candidate_queue->obj_arr[ top].key < dist_k[ i])
				break;
		}

		if( i == query_point_n)	
			break;				//Pruning.

		node_c = candidate_queue->obj_arr[ top].node_v;

		//Check whether node_c is dominated by at least k pruning points;
		if( node_c->parent == NULL)
			MBC_cur = MBC;
		else
			MBC_cur = node_c->parent->MBC_array[ node_c->loc];

		//
		res = is_dom_k( MBC_cur, pruning_list, s_q, k, dom_tag);

		//t
		if( res == 1)
			stat_v_AkNN.AkNN_hn_MBC[ dom_tag] ++;
		//t
			
		if( res != 1)
		{
			//node_c cannot be pruned;
			if( node_c->level > 0)
			{
				//node_c is an inner node;
				//Enqueue all children of node_c;
				for( i=0; i<node_c->num; i++)
				{
					candidate_queue->obj_arr[ cnt].node_v = ( ( node_t*)node_c->child[ i]);
					candidate_queue->obj_arr[ cnt].key = calc_min_min_dist( node_c->MBC_array[ i], s_q);

					insert_b_heap( candidate_queue, cnt++);
				}
			}
			else
			{
				//node_c is a leaf node;
				//Traverse all points under node_c;
				for( i=0; i<node_c->num; i++)
				{
					MBC_cur = ( ( obj_t*)node_c->child[ i])->MBC;

					//Check whether MBC_cur could be pruned;
					res = is_dom_k( MBC_cur, pruning_list, s_q, k, dom_tag);

					//t
					if( res == 1)
						stat_v_AkNN.AkNN_hn_point[ dom_tag] ++;
					//t

					if( res != 1)
					{
						//The point (which is represented by a sphere) cannot be pruned;
						//Enqueue the point into the solution_queue of each query point;

						for( j=0; j<query_point_n; j++)
						{
							//
							//input: MBC_cur, solution_queue[ j], 
							dist_v = calc_dist( MBC_cur->center, R->point_array[ j]);

							if( dist_v < dist_k[ j])
							{
								//
								if( solution_queue[ j]->rear == k)
									cnt_2[ j] = get_top( solution_queue[ j]);
								
								solution_queue[ j]->obj_arr[ cnt_2[ j]].point_v = MBC_cur->center;
								solution_queue[ j]->obj_arr[ cnt_2[ j]].point_v->id = MBC_cur->id;
								solution_queue[ j]->obj_arr[ cnt_2[ j]].key = - dist_v;

								insert_b_heap( solution_queue[ j], cnt_2[ j]);

								if( solution_queue[ j]->rear < k)
									cnt_2[ j] ++;

								//Update the "dist_k"
								dist_k[ j] = - solution_queue[ j]->obj_arr[ solution_queue[ j]->h_arr[ 1]].key;
							}//if
						}//for( query points)
					}//if( res)
				}//for( obj)
			}//else
		}//if( res)
	}//while

	//Construct the solution set;
	for( i=0; i<query_point_n; i++)
		solution_set[ i] = const_point_set_queue( solution_queue[ i]);

	//Release the resource;
	release_point_set_light( pruning_list);
	release_b_heap( candidate_queue);
	release_sphere( MBC);
	for( i=0; i<query_point_n; i++)
		release_b_heap( solution_queue[ i]);
	free( solution_queue);
	free( cnt_2);
	free( dist_k);

	free_SSTree( );

	return solution_set;
}


point_set_t* comp_pruning_list( float eps, int k, sphere_t* s_q)
{	
	//1. It is assumed that the SS_Tree has been constructed already;
	//2. 

	int i, cnt, cnt_2, top;
	DIST_TYPE dist_k, dist_v;

	node_t* node_c;
	obj_t* obj_cur;
	point_set_t* pruning_list;

	b_heap_t* b_h, *b_h_2;

	k = ( int) ( eps * k);

	//Initialize the data structures;
	b_h = alloc_b_heap( 2 * SS_Tree_v.obj_n + 1);

	cnt = 1;
	b_h->obj_arr[ cnt].node_v = SS_Tree_v.root;
	b_h->obj_arr[ cnt].key = 0;

	insert_b_heap( b_h, cnt++);
	
	b_h_2 = alloc_b_heap( k + 1);
	dist_k = DBL_MAX;
	cnt_2 = 1;

	//Perform the "best-first" search;
	while( b_h->rear > 0)
	{
		top = get_top( b_h);

		if( b_h->obj_arr[ top].key >= dist_k)
			break;

		node_c = b_h->obj_arr[ top].node_v;
		
		//
		if( node_c->level > 0)
		{
			//
			for( i=0; i<node_c->num; i++)
			{
				b_h->obj_arr[ cnt].node_v = ( ( node_t*) node_c->child[ i]);
				b_h->obj_arr[ cnt].key = calc_min_min_dist( node_c->MBC_array[ i], s_q);

				insert_b_heap( b_h, cnt++);
			}
		}
		else
		{
			//
			for( i=0; i<node_c->num; i++)
			{
				obj_cur = ( ( obj_t*)node_c->child[ i]);
				dist_v = calc_min_min_dist( obj_cur->MBC, s_q);

				if( dist_v < dist_k)
				{
					//
					if( b_h_2->rear == k)
						cnt_2 = get_top( b_h_2);

					b_h_2->obj_arr[ cnt_2].point_v = obj_cur->MBC->center;
					b_h_2->obj_arr[ cnt_2].point_v->id = obj_cur->MBC->id;
					b_h_2->obj_arr[ cnt_2].key = - dist_v;

					insert_b_heap( b_h_2, cnt_2);

					if( b_h_2->rear < k)
						cnt_2 ++;

					//Update the "best-known"
					if( b_h_2->rear == k)
						dist_k = - b_h_2->obj_arr[ b_h_2->h_arr[ 1]].key;
				}//if
			}//for
		}//else( level == 0)
	}//while

	//Construct the result set
	pruning_list = const_point_set_queue( b_h_2);

	//Release the resources;
	release_b_heap( b_h);
	release_b_heap( b_h_2);

	return pruning_list;
}


int is_dom_k( sphere_t* s_a, point_set_t* point_set_v, sphere_t* s_q, int k, int dom_tag)
{
	int i, cnt;

	sphere_t* s_b;

	cnt = 0;
	for( i=0; i<point_set_v->point_n; i++)
	{
		s_b = const_sphere_point( point_set_v->point_array[ i]);

		if( comp_sphere_dom( s_b, s_a, s_q, dom_tag) == 1)
			cnt ++;

		if( cnt >= k)
			break;

		destroy_sphere_point( s_b);
	}

	if( cnt >= k)
		return 1;
	else
		return 0;
}


point_set_t* const_point_set_queue( b_heap_t* b_h)
{
	int i, size, top;
	point_set_t* point_set_v;

	size = b_h->rear;
	point_set_v = alloc_point_set( SS_Tree_v.dim, size);
	for( i=size-1; i>=0; i--)
	{
		top = get_top( b_h);

		point_set_v->point_array[ i] = b_h->obj_arr[ top].point_v;
	}
	
	return point_set_v;
}

void emp_AkNN( )
{
	int i, f_tag, h, ins_n;
	config_t* config_v;
	sphere_data_t* S;
	point_set_t* R;

	FILE* s_fp;

	//
	memset( &stat_v_AkNN, 0, sizeof( stat_t));
	f_tag = 0;

	//Read the hyperspheres;
	config_v = read_config( );

	S = read_sphere_data( config_v);

	//Change all the hyperspheres to be points;
	set_sphere_radius_zero( S);

	ins_n = 0;
	for( h=0; h<config_v->query_n; h++)
	{
		ins_n ++;
		//if( ins_n % 100 == 0)
			printf( "Instance %i...\n", ins_n);		

		//Generate the query set R;
		R = gen_point_set_v2( S, config_v->query_point_n);
		//R = gen_point_set_v3( S, config_v->query_point_n);
		
		//t
		//print_point_set( stdout, R, 1);
		//t
		
		//Perform the AkNN query;
		for( i=0; i<5; i++)
			AkNN_ins( S, R, config_v->eps, config_v->k, config_v->query_point_n, i);

		//
		release_point_set( R);
	}

	//Print the statistics;
	if( ( s_fp = fopen( STAT_FILE, "w")) == NULL)
	{
		fprintf( stderr, "Cannot open the stat.txt file.\n");
		exit( 0);
	}

	//t
	for( i=0; i<5; i++)
	{
		fprintf( s_fp, "%f\n%f\n%f\n%f\n\n", stat_v_AkNN.AkNN_time[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_pa[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_hn_MBC[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_hn_point[ i] / config_v->query_n);
	}
	
	printf( "Time\tPage access\tHit number(MBC)\t Hit number(point)\n");
	for( i=0; i<5; i++)
		fprintf( stdout, "%f\t%f\t%f\t%f\t\n", stat_v_AkNN.AkNN_time[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_pa[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_hn_MBC[ i] / config_v->query_n, 
										stat_v_AkNN.AkNN_hn_point[ i] / config_v->query_n);

	//Release the resources;
	//release_point_set( R);
	release_sphere_data( S);
	free( config_v);
	fclose( s_fp);
}

void AkNN_ins( sphere_data_t* S, point_set_t* R, float eps, int k, int query_point_n, int dom_tag)
{
	int i;
	point_set_t** solution_set;

#ifndef WIN32
	float usr_t;
	struct rusage start_t, end_t;
		
	get_cur_time( &start_t);
#endif

	solution_set = AkNN( S, R, eps, k, dom_tag);
	
#ifndef WIN32
	get_cur_time( &end_t);
	
	collect_time( &start_t, &end_t, &usr_t);
	stat_v_AkNN.AkNN_time[ dom_tag] += usr_t;
#endif
	
	//t
	//print_AkNN_result( solution_set, config_v->query_point_n, f_tag, HYPER_AKNN_FILE);
	
	//
	for( i=0; i<query_point_n; i++)
		release_point_set_light( solution_set[ i]);
}

void print_AkNN_result( point_set_t** solution_set, int size, int f_tag, const char* f_name)
{
	int i;
	FILE* o_fp;

	//
	if( f_tag == 0)
		o_fp = stdout;
	else
	{
		if( ( o_fp = fopen( f_name, "w")) == NULL)
		{
			fprintf( stderr, "Cannot open %s [print_AkNN_result].\n", f_name);
			exit( 0);
		}
	}

	//
	for( i=0; i<size; i++)
	{
		//
		print_point_set( o_fp, solution_set[ i], 0);
	}

	printf( "\n");

	return;
}