
#ifndef KNN_H
#define KNN_H

#include "sstree.h"
#include "hyperbola.h"
//#include "GP.h"
#include "trigonometric.h"
#include "MBR.h"
#include "MaxMin.h"
#include "bf.h"

#include "b_heap.h"


#define HYPER_AKNN_FILE			"Hyper_AkNN_result.txt"
#define MAXMIN_AKNN_FILE		"Hyper_MAXMIN_result.txt"
#define MBR_AKNN_FILE			"MBR_AkNN_result.txt"
#define GP_AKNN_FILE			"GP_AkNN_result.txt"
#define TRIGONOMETRIC_AKNN_FILE	"Trigo_AkNN_result.txt"


extern stat_t stat_v_DS;
extern stat_t stat_v_AkNN;

int comp_sphere_dom( sphere_t* s_a, sphere_t* s_b, sphere_t* s_q, int dom_opt);

//Application 1: basic "dominating set" and "dominated-by set";
/*
 *	Compute the "dominance set" and the "dominated-by set".
 *
 *	@prob_tag =
 *				1: "dominance set"
 *				2: "dominated-by set"
 *	@alg_tag =
 *				1: Linear-Scan
 *				2: index-based
 *	
 *	@dom_tag: indicates the pruning techniques;
 *				
 */
sphere_list_t* comp_basic_DS( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int alg_opt, int dom_opt);

/*
 *	Linear-Scan the spheres in the data set;
 *	Complexity: O(N d), where N is the number of spheres in the data set.
 */
sphere_list_t* comp_basic_DS_linear_scan( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int dom_opt);

/*
 *	The implementation based on the SSTree;
 *	
 *	
 */
sphere_list_t* comp_basic_DS_index( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int prob_tag, int dom_opt);

//sphere_list_t* comp_basic_DBS( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int alg_opt, int dom_opt);

//sphere_list_t* comp_basic_DBS_linear_scan( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int dom_opt);

//sphere_list_t* comp_basic_DBS_index( sphere_data_t* sphere_data_v, sphere_t* s_a, sphere_t* s_q, int dom_opt);


void emp_basic_DB( );


//Application 2: AkNN queries;
	//1. data
		//- A set S of points
		//- A set R of query points
	
	//s. parameters:
		//- k
		//- \eps
		//- |R|

	//3. 

sphere_t* const_sphere_point( point_t* point_v);

void destroy_sphere_point( sphere_t* sphere_v);

point_set_t** AkNN( sphere_data_t* S, point_set_t* R, float eps, int k, int dom_tag);

void AkNN_ins( sphere_data_t* S, point_set_t* R, float eps, int k, int query_point_n, int dom_tag);

point_set_t* comp_pruning_list( float eps, int k, sphere_t* s_q);

int is_dom_k( sphere_t* s_a, point_set_t* point_set_v, sphere_t* s_q, int k, int dom_tag);

point_set_t* const_point_set_queue( b_heap_t* b_h);

void print_AkNN_result( point_set_t** solution_set, int size, int f_tag, const char* f_name);

//
void emp_AkNN( );


#endif