
#include "sstree.h"

/*----------------------Calculation related APIs--------------------------*/

/*
* abs is over-loading edition of 'abs(int)' in math.h.
*
*  Same.
* 
* return the absolute MBR_array of var(double).
**
double abs(double var)
{
	if(var<0)
		return -var;
	else 
		return var;
};
*/

int compare_f (const void * a, const void * b)
{
  if( *( float*)a - *( float*)b > 0)
	  return 1;
  else if( *( float*)a - *( float*)b < 0)
	  return -1;
  else 
	  return 0;
}

/*
* CreateNode malloc the storage space for a pointer to 'node'.
*
*  pnode indicates the pointer.
*
* return void.
*/
void CreateNode( node_t* &pnode)
{
	pnode = ( node_t*)malloc( sizeof( node));
	memset( pnode, 0, sizeof( node));

	pnode->child = ( void**)malloc( ( M+1) * sizeof( void*));
	memset( pnode->child, 0, ( M + 1) * sizeof( void*));
	pnode->MBR_array = ( _range**)malloc( ( M+1) * sizeof( _range*));
	memset( pnode->MBR_array, 0, ( M + 1) * sizeof( _range*));

	/*s*/
	float tmp = _stat_v.memory_v;
	/*s*/

	//IR-tree augmentation.
	//pnode->bst_v = bst_ini( );

	/*s*/
	_stat_v.tree_memory_v += _stat_v.memory_v - tmp;
	_stat_v.memory_v = tmp;

	_stat_v.tree_memory_v += sizeof( node) + ( M + 1) * ( sizeof( void*) + sizeof( _range*));
	if( _stat_v.tree_memory_v > _stat_v.tree_memory_max)
		_stat_v.tree_memory_max = _stat_v.tree_memory_v;
	/*s*/

/*	
	for(i=0; i<M; i++)
	{
		pnode->child[i] = (node_t*)malloc(sizeof(node));
		pnode->MBR_array[i] = (_range*)malloc(sizeof(_range));
	}
*/
}

void CreateNode_ss( node_t* &pnode)
{
	pnode = ( node_t*)malloc( sizeof( node));
	memset( pnode, 0, sizeof( node));

	pnode->child = ( void**)malloc( ( M+1) * sizeof( void*));
	memset( pnode->child, 0, ( M + 1) * sizeof( void*));

	pnode->MBC_array = ( sphere_t**)malloc( ( M+1) * sizeof( sphere_t*));
	memset( pnode->MBC_array, 0, ( M + 1) * sizeof( sphere_t*));

	/*s*/
	float tmp = _stat_v.memory_v;
	/*s*/

	//IR-tree augmentation.
	//pnode->bst_v = bst_ini( );

	/*s*/
	_stat_v.tree_memory_v += _stat_v.memory_v - tmp;
	_stat_v.memory_v = tmp;

	_stat_v.tree_memory_v += sizeof( node) + ( M + 1) * ( sizeof( void*) + sizeof( sphere_t*));
	if( _stat_v.tree_memory_v > _stat_v.tree_memory_max)
		_stat_v.tree_memory_max = _stat_v.tree_memory_v;
	/*s*/

/*	
	for(i=0; i<M; i++)
	{
		pnode->child[i] = (node_t*)malloc(sizeof(node));
		pnode->MBR_array[i] = (_range*)malloc(sizeof(_range));
	}
*/
}

/*
 *	Release a node structure.
 */
void ReleaseNode( node_t* pnode, int tag)
{
	int i;

	if( tag && pnode->level > 0)
	{
		for( i=0; i<pnode->num; i++)
			free( pnode->MBR_array[ i]);

		/*s*/
		_stat_v.tree_memory_v -= SS_Tree_v.dim * sizeof( _range);
		/*s*/
	}

	free( pnode->child);
	free( pnode->MBR_array);

	/*s*/
	float tmp = _stat_v.memory_v;
	/*s*/

	//bst_release( pnode->bst_v);

	/*s*/
	_stat_v.tree_memory_v -= tmp - _stat_v.memory_v;
	_stat_v.memory_v = tmp;
	/*s*/

	free( pnode);

	/*s*/
	_stat_v.tree_memory_v -= sizeof( node) + ( M + 1) * ( sizeof( void*) + sizeof( _range*));
	/*s*/
}

void ReleaseNode_ss( node_t* pnode, int tag)
{
	int i;

	if( tag && pnode->level > 0)
	{
		for( i=0; i<pnode->num; i++)
			//free( pnode->MBR_array[ i]);
			release_sphere( pnode->MBC_array[ i]);

			//Need to update the memory statistics for "release_sphere".
	}

	free( pnode->child);
	free( pnode->MBC_array);
	free( pnode);
}

/*
* IsOverlapped determines whether MBR overlaps the obj.
*
* param MBR indicates the MBRangle.
* param obj is the object.
*
* return 1 if overlapped; 0 if crossed; else(departed) -1.
*/
int IsOverlapped(_range* MBR, _range* obj)
{
	int  i, tag;
	tag = 0;
	for( i=0; i<SS_Tree_v.dim; i++)	//Travel each dimension.
	{
		//Departed;
		if( MBR[ i].min > obj[ i].max ||
			MBR[ i].max < obj[ i].min)
			return -1;
		//Crossed;
		else if( !( MBR[ i].min >= obj[ i].min &&
			MBR[ i].max <= obj[ i].max))
			tag = 1;
		//Contained;
		else 
			continue;
	}
	if( tag = 1)
		return 0;
	else
		return 1;
}

/*
* IsSame is used to judge whether the two MBRangles are the same.
*
*  MBR1, MBR2 indicate the two MBRangles respectively.
*
* return true if same, else false.
*/
bool IsSame( _range* MBR1, _range* MBR2)
{
	int i;
	for( i=0; i<SS_Tree_v.dim; i++)
	{
		if( !(MBR1[i].max == MBR2[i].max && 
			MBR1[i].min == MBR2[i].min))
			return false;
	}
	return true;
}

/*
* calc_MBR is used to find the closing MBRangle for a node.
*
*  pnode indicates the node in rtree.
*  MBR gets back the closing MBRangle(allocated outside the function).
* 
* return void.
*/
void calc_MBR( node_t* pnode, _range* &MBR)
{
	//The interface of "MBR" is due to the R-tree implementation.
	int i, j;
	for( j=0; j<SS_Tree_v.dim; j++)         //dimensions.
	{
		MBR[j].min = FLT_MAX;
		MBR[j].max = -FLT_MAX;

		for( i=0; i<pnode->num; i++)      //entries.
		{
			if( pnode->MBR_array[i][j].min < MBR[j].min)
				MBR[j].min = pnode->MBR_array[i][j].min;
			if( pnode->MBR_array[i][j].max > MBR[j].max)
				MBR[j].max = pnode->MBR_array[i][j].max;
		}
	}
}

/*
 *	@opt:
 *		= 0: the MBC has not been allocated (thus it should be allocated in this funciton);
 *		= 1: the MBC has already been allocated.
 */
void calc_MBC_ss( node_t* pnode, sphere_t* &MBC, int opt)
{
	int i, j, dim;
	DIST_TYPE dist_v;

	dim = SS_Tree_v.dim;

	if( opt == 0)
		MBC = alloc_sphere( dim);
	else
		ini_sphere( MBC);

	//Specify the center;
	for( i=0; i<dim; i++)
	{
		for( j=0; j<pnode->num; j++)
			MBC->center->coord[ i] += pnode->MBC_array[ j]->center->coord[ i];

		MBC->center->coord[ i] /= dim;
	}

	//Specify the radius;
	MBC->radius = 0;
	for( i=0; i<pnode->num; i++)
	{
		dist_v = calc_dist( MBC->center, pnode->MBC_array[ i]->center) + pnode->MBC_array[ i]->radius;

		if( dist_v > MBC->radius)
			MBC->radius = dist_v;
	}

	return;
}


/*
 *	Same usage as calc_MBR with a different interface.
 */
_range* get_MBR_node( node_t* pnode, int dim)
{
	_range* MBR;

	MBR = ( _range*)malloc( dim * sizeof( _range));
	memset( MBR, 0, dim * sizeof( _range));

	calc_MBR( pnode, MBR);

	return MBR;
}

/*
* GetArea is used to get the area of closing MBRangle of a node.
*
*  pnode indicates the node.
*
* return the area MBR_array.
**/
double GetArea( node_t* pnode)
{
	int i;
	double area;
	_range* MBR;

	MBR = get_MBR_node( pnode, SS_Tree_v.dim);

	area = 1;
	for(i=0; i<SS_Tree_v.dim; i++)
		area *= MBR[i].max - MBR[i].min;

	free( MBR);
	return area;
}

/*
* GetArea_2 is used to get the closing area of two MBRangles.
* 
*  MBR1, MBR2 indicates two MBRangles respectively.
*
* return the area.
**/ 
double GetArea_2( _range* MBR1, _range* MBR2)
{
	double area;
	float high, low;
	int i;

	area = 1;
	for( i=0; i<SS_Tree_v.dim; i++)
	{	
		high = MBR1[i].max >= MBR2[i].max ? MBR1[i].max : MBR2[i].max;
		low = MBR1[i].min <= MBR2[i].min ? MBR1[i].min : MBR2[i].min;
		
		area *= high - low;
	}

	return area;
}

/*
* GetIncArea is used to get the increased area when adding a MBRangle into a node.
*
*  pnode indicates the node.
*  MBR indicates the MBRangle.
*
* return the increased area.
**/
double GetIncArea( node_t* pnode, _range* MBR)
{
	double area1, area2;
	
	//Original Area in pnode.
	area1 = GetArea( pnode);

	//Area after adding MBR into node_1.
	pnode->MBR_array[ pnode->num] = MBR;
	pnode->num ++;
	area2 = GetArea( pnode);
	pnode->num --;
	
	return area2 - area1;
}

/*
*This funtion is used to locate the closing MBRangle 
*which need the least enlargement in order to enclose the obj.
*
*  pnode indicates the node in the rtree.
*  obj indicates the object.
*  loc get back the location of subtree to be propagate.
*
* return void.
**/
void NeedLeastArea( node_t* pnode, _range* obj, int& loc)
{
	int i, j;
	double ori, enl, pre_area, min_inc_area;
	_range* MBR, *MBR_tmp;

	//use MBR_tmp to denote the enlarged MBRangle.
	MBR_tmp = ( _range*)malloc( SS_Tree_v.dim * sizeof( _range));
	memset( MBR_tmp, 0, SS_Tree_v.dim * sizeof( _range));

	loc = 0;
	min_inc_area = DBL_MAX;
	for(i=0; i<pnode->num; i++)            //entries
	{
		ori = 1;
		enl = 1;
		MBR = pnode->MBR_array[i];
		for( j=0; j<SS_Tree_v.dim; j++)              //dimensions
		{
			MBR_tmp[ j].min = MBR[ j].min <= obj[ j].min ?
				MBR[ j].min : obj[ j].min;
			MBR_tmp[ j].max = MBR[ j].max >= obj[ j].max ?
				MBR[ j].max : obj[ j].max;

			ori *= MBR[ j].max - MBR[ j].min;
			enl *= MBR_tmp[ j].max - MBR_tmp[ j].min;
		}//for(j)
		
		if( min_inc_area > enl-ori)
		{
			min_inc_area = enl - ori;
			loc = i;
			pre_area = ori;
		}
		else if( min_inc_area == enl-ori)
		{
			if( ori < pre_area)
			{
				pre_area = ori;
				loc = i;
			}
		}
	}//for(i)

	free(MBR_tmp);

	return;
}

void NeedLeastArea_ss( node_t* pnode, sphere_t* MBC, int& loc)
{
	int i;
	double ori, enl, pre_area, min_inc_area;

	//
	sphere_t* MBC_iter;

	loc = 0;
	min_inc_area = DBL_MAX;
	for(i=0; i<pnode->num; i++)            //entries
	{
		//MBR = pnode->MBR_array[i];
		MBC_iter = pnode->MBC_array[ i];

		//Calculate the area difference.
		ori = PI * 2 * MBC_iter->radius;
		enl = PI * ( MBC_iter->radius + calc_dist( MBC_iter->center, MBC->center), MBC->radius);

		if( min_inc_area > enl-ori)
		{
			min_inc_area = enl - ori;
			loc = i;
			pre_area = ori;
		}
		else if( min_inc_area == enl - ori)
		{
			if( ori < pre_area)
			{
				pre_area = ori;
				loc = i;
			}
		}
	}//for(i)

	return;
}

/*
* UpdateMBR is used to find the closing MBRangle & judgement of enlargement.
*
*  same with 'calc_MBR's.
*
* return true if enlarged, else false.
*/
bool UpdateMBR( node_t* pnode, _range* &MBR)
{
	int i, j;
	float min, max;
	bool res;

	res = false;
	for( i=0; i<SS_Tree_v.dim; i++)		//dimensions.
	{
		min = FLT_MAX;
		max = -FLT_MAX;

		for(j=0; j<pnode->num; j++)		//entries.
		{
			if( pnode->MBR_array[j][i].min < min)
				min = pnode->MBR_array[j][i].min;
			if( pnode->MBR_array[j][i].max > max)
				max = pnode->MBR_array[j][i].max;
		}

		//Be careful of the 'initial' conditions of MBR--bug (bigger OR smaller).
		if(MBR[i].max != max)
		{
			MBR[i].max = max;
			res = true;
		}
		if(MBR[i].min != min)
		{
			MBR[i].min = min;
			res = true;
		}
	}

	return res;
}

bool UpdateMBC_ss( node_t* pnode, sphere_t* &MBC)
{
	int i;
	DIST_TYPE dist_v;
	bool res;

	res = false;
	for( i=0; i<pnode->num; i++)
	{
		dist_v = calc_dist( pnode->MBC_array[ i]->center, MBC->center) + pnode->MBC_array[ i]->radius;
		
		if( dist_v > MBC->radius)
		{
			res = true;
			break;
		}
	}

	if( res == true)
	{
		calc_MBC_ss( pnode, MBC, 1);
	}

	return res;
}  

/*------------------------Insertion related APIs-------------------------*/


int find_closest_branch( node_t* node_v, sphere_t* MBC)
{
	int i, branch_no;
	DIST_TYPE dist_v, dist_min;

	branch_no = 0;
	dist_min = DBL_MAX;
	for( i=0; i<node_v->num; i++)
	{
		dist_v = calc_dist( node_v->MBC_array[ i]->center, MBC->center);

		if( dist_v < dist_min)
		{
			dist_min = dist_v;
			branch_no = i;
		}
	}

	return branch_no;
}

/*
*This function is used to choose a leaf node to include the obj_t to be inserted.
*
*  root indicates the rtree.
*  MBR indicates the object for which to choose a leaf-node.
*  leaf_node is used to get back to objective leaf node.
*
* return void.
*/
void ChooseLeaf( node_t* root, _range* MBR, node_t* &leaf_node)
{
	int loc = 0;
	while(1)
	{	
		if(root->level==0)
		{
			//root node is a leaf-node.
			leaf_node = root;
			return;
		}		

		NeedLeastArea( root, MBR, loc);
		root = ( node_t*)( root->child[loc]);
	}
}


void ChooseLeaf_ss( node_t* root, sphere_t* MBC, node_t* &leaf_node)
{
	int loc;
	
	loc = 0;

	while( 1)
	{	
		if( root->level == 0)
		{
			//root node is a leaf-node.
			leaf_node = root;

			return;
		}		

		//NeedLeastArea( root, MBR, loc);
		//NeedLeastArea_ss( root, MBC, loc);
		//Choose the branch that has the center closest to the center of the object.
		loc = find_closest_branch( root, MBC);
		
		root = ( node_t*)( root->child[ loc]);
	}
}

/*
*This function is used to pick up two seed MBRangles.
*
*  pnode indicates the node to be split.
*  MBR is the object to be inserted.
*  g1, g2 get back the locations of the seeds( g1 = -1 for the MBR)
* 
* return void.
*/
void PickSeeds( node_t* pnode, _range* MBR, int& g1, int& g2)
{
	int i, j, k;
	double area, area1, area2, max = -DBL_MAX;
	_range* MBR1, *MBR2, *MBR_tmp;
	
	MBR_tmp = ( _range*)malloc( sizeof(_range));
	memset( MBR_tmp, 0, sizeof( _range));
	
	MBR1 = MBR;
	for(i=-1; i<pnode->num; )			
	{
		for(j=0; j<pnode->num; j++)	
		{
			MBR2 = pnode->MBR_array[j];

			if( j != i)
			{
				area = 1;
				area1 = 1;
				area2 = 1;

				for(k=0; k<SS_Tree_v.dim; k++)
				{	
					//MBR_tmp is used to enclose the two MBR_array: MBR1, MBR2.
					MBR_tmp->max = (MBR1[k].max >= MBR2[k].max)? MBR1[k].max : MBR2[k].max;
					MBR_tmp->min = (MBR1[k].min <= MBR2[k].min)? MBR1[k].min : MBR2[k].min;
					area *= MBR_tmp->max - MBR_tmp->min;
					area1 *= MBR1[k].max - MBR1[k].min;
					area2 *= MBR2[k].max - MBR2[k].min;
				}//
				
				if( area - area1 - area2 > max)
				{
					max = area - area1 - area2;
					g1 = i;
					g2 = j;
				}
			}//if(j!=i)
		}//if(j)

		MBR1 = pnode->MBR_array[ ++i];
	}//if(i)
	
	free( MBR_tmp);
}

/*
* LinearPickSeeds is used to pick the seeds for 'LinearSplit' function.
* 
* With the same interface as 'PickSeeds'.
*/
void LinearPickSeeds( node_t* pnode, _range* MBR, int& g1, int& g2)
{
	//double aTmp, maxArea = -1;
	float lowHighSide, highLowSide, width, sep, sep_max;
	int i, j, gTmp1, gTmp2;
	_range *rTmp, *MBR0;

	//Get the MBR of pnode, which is used for normalizing the separation.
	MBR0 = ( _range*)malloc( SS_Tree_v.dim * sizeof( _range));
	memset( MBR0, 0, sizeof( SS_Tree_v.dim * sizeof( _range)));

	calc_MBR( pnode, MBR0);
	
	//
	sep_max = 0;
	for( i=0; i<SS_Tree_v.dim; i++)
	{	
		//Travel each dimension.
		lowHighSide = FLT_MAX;
		highLowSide = -FLT_MAX;

		//Find two MB_ranges with the "highest lower side" and 
		//the "lowest higher side", respectively.
		rTmp = MBR;
		for( j=-1; j<pnode->num; )
		{
			if( rTmp[i].max < lowHighSide)
			{
				gTmp1 = j;
				lowHighSide = rTmp[i].max;
			}

			if( rTmp[i].min > highLowSide)
			{
				gTmp2 = j;
				highLowSide = rTmp[i].min;
			}

			rTmp = pnode->MBR_array[ ++j];
		}
		
		if( gTmp1 == gTmp2) 
		{
			//The two entries correspond to the same MBR;
			//Randomly pick another entry.
			if( gTmp1 < pnode->num-1)
				gTmp2 = gTmp1 + 1;
			else
				gTmp2 = gTmp1 - 1;
		}

		//Calculate the sepration between MBR1 and MBR2 along dim i.
		sep = max( highLowSide, lowHighSide) -
			  min( highLowSide, lowHighSide);

		//Recloc the normalized separation.
		width = max( MBR0[ i].max, MBR[ i].max) - 
				min( MBR0[ i].min, MBR[ i].min);
		sep /= width;

		if( sep > sep_max)
		{
			sep_max = sep;
			g1 = gTmp1;
			g2 = gTmp2;
		}
		
		//No tie handling for sep == sep_max here.		
	}

	//Release the resource.
	free( MBR0);

	return;
}

/*
* PickNext is used to pick the new item to be grouped.
* 
*  pnode is the node to split.
*  node_1, node_2 indicate the two split nodes.
*  node_c stores the node to which the 'next' entry is assigned.
*  MBR indicates the MBR of the object/sub-tree to be inserted (valid when rTag == 0).
*  done indicates the entries that have been assigned.
*
* return the loc of the next entry.
*/
int PickNext( node_t* pnode, node_t* node_1, node_t* node_2, node_t* &node_c, _range* MBR, bool* done, int rTag)
{	
	//Code is long, but the effiency is high.
	//It could be shorten, but it's better not.
	//Tie handling: Norbert's "R*-tree".
	int i, next;
	double max = -DBL_MAX, d1, d2, dif;
	_range* MBR_tmp;

	if( rTag == 0)
	{
		i = -1;
		MBR_tmp = MBR;
	}
	else
	{
		i = 0;
		MBR_tmp = pnode->MBR_array[ i];
	}

	for( ; i<pnode->num; )
	{	
		//Travel all the entries in pnode that haven't been grouped.
		if( !( i>= 0 && done[ i] == true))
		{				
			d1 = GetIncArea( node_1, MBR_tmp);
			d2 = GetIncArea( node_2, MBR_tmp);			
			dif = d2 - d1;	
			
			if( fabs( dif) > max )
			{
				max = fabs( dif);
				next = i;

				//Decide the node.
				if(dif > 0)
					node_c = node_1;
				else if( dif == 0)		
				{
					d1 = GetArea( node_1);
					d2 = GetArea( node_2);
					if( d1 < d2)
						node_c = node_1;
					else if( d1 == d2)
					{
						if( node_1->num <= node_2->num)
							node_c = node_1;
						else
							node_c = node_2;
					}
					else
						node_c = node_2;
				}
				else
					node_c = node_2;
			}
		}

		i++;
		MBR_tmp = pnode->MBR_array[ i];
	}

	return next;
}

/*
* LinearPickNext has the same function as 'PickNext' except that it picks the next entry randomly.
*
* The same interface as "PickNext".
*
*/
int LinearPickNext( node_t* pnode, node_t* node_1, node_t* node_2, node_t* &node_c, bool* done, int rTag)
{
	int i, next;

	//
	if( rTag == 0)
	{
		next = -1;
	}
	else
	{
		for(i=0; i<pnode->num; i++)
		{
			if( done[i] == false)
			{
				next = i;
				break;
			}
		}
	}//else
	
	//
	if( node_1->num > node_2->num)
		node_c = node_1;
	else
		node_c = node_2;
	
	return next;
}

/*
 *	Assign an entry to a node.
 *
 *	pnode involves the 'level' info.
 *	node_c indicates the node to which the entry is assigned.
 *	( obj, MBR) indicates the entry.
 */
void AssignEntry( node_t* pnode, node_t* node_c, void* obj, _range* MBR)
{
	node_c->child[ node_c->num] = obj;
	node_c->MBR_array[ node_c->num] = MBR;
	
	if( pnode->level > 0)
	{
		( ( node_t*)obj)->parent = node_c;
		( ( node_t*)obj)->loc = node_c->num;
	}

	node_c->num ++;
}

void AssignEntry_ss( node_t* pnode, node_t* node_c, void* obj, sphere_t* MBC)
{
	node_c->child[ node_c->num] = obj;
	//node_c->MBR_array[ node_c->num] = MBR;
	node_c->MBC_array[ node_c->num] = MBC;
	
	if( pnode->level > 0)
	{
		( ( node_t*)obj)->parent = node_c;
		( ( node_t*)obj)->loc = node_c->num;
	}

	node_c->num ++;
}

/*
* SplitNode is used to split the node into two nodes.
*
*  pnode is the node to be split--Note the storage space before invoke.
*  node_2 is the new node allocated outside the function to store the new node.
*  obj indicates the object/sub-tree to be inserted.
*  opt = 1 for LinearSplit and opt = 2 for QuadraticSplit.
*
* return void.
*/
void SplitNode( node_t* &pnode, node_t* &node_2, void* obj, int opt)
{   
	//Note that pnode pointer has been updated in this function.
	//It is required to update the child attribute of pnode's parent if it exists.

	int i, g1, g2, rTag = 0, qCase, next = 0;
	node_t* node_c, *node_1;

	//
	CreateNode( node_1);
	CreateNode( node_2);
	
	//done marks the elements in pnode that have been grouped.
	bool* done = (bool*)malloc( sizeof( bool) * M);
	memset( done, 0, sizeof( bool) * M);

	//PickSeeds.
	_range* MBR;
	if( pnode->level > 0)
	{	
		//The inserted is a subtree.
		MBR = ( _range*)malloc( SS_Tree_v.dim * sizeof( _range));
		memset( MBR, 0, SS_Tree_v.dim * sizeof( _range));

		calc_MBR( ( node_t*)obj, MBR);
	}
	else
	{
		//The inserted is an object.
		MBR =( ( obj_t*)obj)->MBR;
	}

	if( opt == 1)
		LinearPickSeeds( pnode, MBR, g1, g2);
	else
		PickSeeds( pnode, MBR, g1, g2);
	
	//Assign the two entries indicated by g1, g2.
	if( g1 == -1)    //obj
	{
		//g1 indicates the obj.
		AssignEntry( pnode, node_2, obj, MBR);
		
		rTag = 1;	//For indicating the object?
	}
	else
	{
		//g1 indicates an entry of pnode.
		AssignEntry( pnode, node_2, pnode->child[ g1], pnode->MBR_array[ g1]);
		done[ g1] = true;
	}

	AssignEntry( pnode, node_1, pnode->child[ g2], pnode->MBR_array[ g2]);
	done[ g2] = true;
	
	//Assign the rest entries.
	while( true)
	{
		if( node_1->num + node_2->num == M + 1 )
		{   
			// 1. All entries have been assigned.
			qCase = 1;
			break;
		}
		else if( node_1->num == M + 1 - m)
		{	
			// 2. The rest should be assigned to node_2.
			qCase = 2;
			break;
		}
		else if( node_2->num == M + 1 - m)
		{	
			// 3. The rest should be assigned to node_1.
			qCase = 3;
			break;
		}

		//PickNext.
		if( opt == 1)
			next = LinearPickNext( pnode, node_1, node_2, node_c, done, rTag);
		else
			next = PickNext( pnode, node_1, node_2, node_c, MBR, done, rTag);

		//Assign the 'next' entry.
		if( next == -1)	//the 'next' is obj.
		{
			AssignEntry( pnode, node_2, obj, MBR);
			rTag = 1;
		}
		else
		{
			AssignEntry( pnode, node_c, pnode->child[ next], pnode->MBR_array[ next]);
			done[ next] = true;
		}
	}//while(1)
	
	//Assign the rest entries if necessary.
	if( qCase == 2)
		node_c = node_2;
	else if( qCase == 3)
		node_c = node_1;
	else
		goto E;

	if( rTag == 0)
	{
		AssignEntry( pnode, node_c, obj, MBR);
		rTag = 1;
	}	
	
	for(i=0; i<pnode->num; i++)
	{
		if( done[i] == false)
		{
			AssignEntry( pnode, node_c, pnode->child[ i], pnode->MBR_array[ i]);
			done[ i] = true;
		}
	}
	
	//
E:	node_1->parent = pnode->parent;	
	node_1->loc = pnode->loc;
	node_1->level = pnode->level;
	//((node_t*)(pnode->parent))->child[pnode->loc] = node_1;
	
	node_2->level = pnode->level;
	//Note that node_2's 'parent' & 'loc' haven't been set.
	
	node_c = pnode;
	pnode = node_1;

	//Free the resource.
	ReleaseNode( node_c, 0);
	free( done);
}

void SplitNode_ss( node_t* &pnode, node_t* &node_2, void* obj)
{
	//Tasks:
		//- split pnode into pnode and node_2;
		//- obj could be an sub-tree or an object;
		//

	int i, dim, size;
	int* group_indicator;
	
	sphere_t* MBC;

	node_t* node_v1, *node_v2, *node_c;

	if( pnode->level > 0)
		//MBC = ( ( node_t*)obj)->MBC;
		calc_MBC_ss( ( ( node_t*)obj), MBC, 0);
	else
		MBC = ( ( obj_t*)obj)->MBC;

	//Step 1: find the dimension with the greatest "variance"
		//- What's the definition of variance *wrt a dimension*
		//- It should be based on the centers' coordinates on this dimension.

	dim = pick_axis_variance_ss( pnode, MBC);

	//Step 2: Split the node;
		//- The objective is to minimized the sum of the variances of the splited nodes.
		//- An option is to use the median of the centers.

	//
	group_indicator = calc_group_indicator_ss( pnode, MBC, dim);
	
	CreateNode_ss( node_v1);
	CreateNode_ss( node_v2);

	size = pnode->num + 1;
	for( i=0; i<size; i++)
	{
		if( group_indicator[ i] == 1)
			node_c = node_v1;
		else
			node_c = node_v2;

		//
		if( i < size-1)
			AssignEntry_ss( pnode, node_c, pnode->child[ i], pnode->MBC_array[ i]);
		else
			AssignEntry_ss( pnode, node_c, obj, MBC);
	}

	//
	node_v1->parent = pnode->parent;
	node_v1->loc = pnode->loc;
	node_v1->level = pnode->level;

	node_v2->level = pnode->level;

	//
	node_c = pnode;
	pnode = node_v1;
	node_2 = node_v2;

	//Release the resources;
	ReleaseNode_ss( node_c, 0);

	//if( pnode->level > 0)
	//	release_sphere( MBC);

	free( group_indicator);
}

double calc_variance_ss( COORD_TYPE* values, int size)
{
	int i;
	COORD_TYPE mean;
	double var;

	mean = 0;
	for( i=0; i<size; i++)
		mean += values[ i];
	mean /= size;

	var = 0;
	for( i=0; i<size; i++)
		var += pow( values[ i] - mean, 2);
	
	return var;
}

COORD_TYPE calc_median_ss( COORD_TYPE* values, int size)
{
	qsort( values, size, sizeof( COORD_TYPE), compare_f);

	return values[ size / 2];
}

int* calc_group_indicator_ss( node_t* pnode, sphere_t* MBC, int dim)
{
	int i, size, cnt, tag;
	int* group_indicator;
	COORD_TYPE median;
	COORD_TYPE* values;

	size = pnode->num + 1;
	values = ( COORD_TYPE*)malloc( size * sizeof( COORD_TYPE));
	memset( values, 0, size * sizeof( COORD_TYPE));

	group_indicator = ( int*)malloc( size * sizeof( int));
	memset( group_indicator, 0, size * sizeof( int));

	//
	for( i=0; i<size-1; i++)
		values[ i] = pnode->MBC_array[ i]->center->coord[ dim];
	values[ i] = MBC->center->coord[ dim];

	median = calc_median_ss( values, size);

	//Assign the entries with their dim^th coordinates at most median to the first node 
	//(At most half of the entries are assigned)
	tag = 0;
	cnt = 0;
	for( i=0; i<size-1; i++)
	{
		if( pnode->MBC_array[ i]->center->coord[ dim] <= median)
		{
			group_indicator[ i] = 1;
			cnt ++;
		}

		if( cnt >= size / 2)
		{
			tag = 1;
			break;
		}
	}

	if( tag != 1)
	{
		if( MBC->center->coord[ dim] <= median)
			group_indicator[ size-1] = 1;
	}
	
	//Assign the remaining entries to the second node.
	for( i=0; i<size; i++)
	{
		if( group_indicator[ i] == 0)
			group_indicator[ i] = 2;
	}

	//Release the resources.
	free( values);
	
	return group_indicator;
}

int pick_axis_variance_ss( node_t* pnode, sphere_t* MBC)
{
	int i, j, size, dim;
	COORD_TYPE* values;
	DIST_TYPE var_max, var_v;

	//
	size = pnode->num + 1;
	values = ( COORD_TYPE*)malloc( size * sizeof( COORD_TYPE));
	memset( values, 0, size * sizeof( COORD_TYPE));

	//	
	dim = 0;
	var_max = 0;
	for( i=0; i<SS_Tree_v.dim; i++)
	{
		for( j=0; j<size-1; j++)
		{
			values[ j] = pnode->MBC_array[ j]->center->coord[ i];
		}
		values[ j] = MBC->center->coord[ i];

		var_v = calc_variance_ss( values, size);

		if( var_v > var_max)
		{
			var_max = var_v;
			dim = i;
		}
	}
	
	//Release the resources.
	free( values);
	
	return dim;
}

/*
 *	Create a new root due to the split operation.
 *	
 *	node_1 is the original root.
 *	node_2 is the new split node from node_1.
 */
node_t* CreateNewRoot( node_t* node_1, node_t* node_2)
{
	int i;
	node_t* nRoot;

	CreateNode( nRoot);
	nRoot->num = 2;
	nRoot->parent = NULL;
	nRoot->level = node_1->level + 1;

	for( i=0; i<2; i++)
	{
		nRoot->child[ i] = ( node_t*)malloc( sizeof(node));
		memset( nRoot->child[ i], 0, sizeof( node));

		nRoot->MBR_array[ i] = ( _range*)malloc( SS_Tree_v.dim * sizeof( _range));
		memset( nRoot->MBR_array[ i], 0, SS_Tree_v.dim * sizeof( _range));
	}

	/*s*/
	_stat_v.tree_memory_v += 2 * ( sizeof( node) + SS_Tree_v.dim * sizeof( _range));
	if( _stat_v.tree_memory_v > _stat_v.tree_memory_max)
		_stat_v.tree_memory_max = _stat_v.tree_memory_v;
	/*s*/

	//Adjust the two children of nRoot: node_1, node_2;
	node_1->parent = nRoot;
	node_1->loc = 0;
	calc_MBR( node_1, nRoot->MBR_array[ 0]);
	nRoot->child[ 0] = node_1;

	node_2->parent = nRoot;
	node_2->loc = 1;
	calc_MBR( node_2, nRoot->MBR_array[ 1]);
	nRoot->child[ 1] = node_2;

	//const_IF( nRoot);

/*t/
	print_IF( nRoot);
/*t*/

	SS_Tree_v.inner_n ++;

	return nRoot;
}

node_t* CreateNewRoot_ss( node_t* node_1, node_t* node_2)
{
	node_t* nRoot;

	CreateNode_ss( nRoot);
	nRoot->num = 2;
	nRoot->parent = NULL;
	nRoot->level = node_1->level + 1;

	nRoot->child[ 0] = node_1;
	calc_MBC_ss( node_1, nRoot->MBC_array[ 0], 0);

	nRoot->child[ 1] = node_2;
	calc_MBC_ss( node_2, nRoot->MBC_array[ 1], 0);

	//Adjust the two children of nRoot: node_1, node_2;
	node_1->parent = nRoot;
	node_1->loc = 0;
	
	node_2->parent = nRoot;
	node_2->loc = 1;
	
	//const_IF( nRoot);

/*t/
	print_IF( nRoot);
/*t*/

	SS_Tree_v.inner_n ++;

	return nRoot;
}

/*
* InsertSub is used to insert an node/subtree recursively.
*
*  pnode indicates the node to insert in.
*  obj indicates the node(sbutree) to be inserted in.
*
* return NULL if the root hasn't changed;
* otherwise, return the new root.
*/
node_t* InsertSub( node_t* pnode, void* obj)
{
	if( pnode->num < M)
	{
		//pnode has room for the new entry.
		if( pnode->level > 0)
		{
			//iner-node case.
			pnode->MBR_array[ pnode->num] = ( _range*)malloc( SS_Tree_v.dim * sizeof( _range));
			memset( pnode->MBR_array[ pnode->num], 0, SS_Tree_v.dim * sizeof( _range));
			
			calc_MBR( ( node_t*)obj, pnode->MBR_array[ pnode->num]);
			( (node_t*)obj)->parent = pnode;
			( (node_t*)obj)->loc = pnode->num;                      //bug.

			/*s*/
			_stat_v.tree_memory_v += SS_Tree_v.dim * sizeof( _range);
			if( _stat_v.tree_memory_v > _stat_v.tree_memory_max)
				_stat_v.tree_memory_max = _stat_v.tree_memory_v;
			/*s*/
		}
		else
		{
			//leaf-node case.
			pnode->MBR_array[ pnode->num] = (( obj_t*)obj)->MBR;
		}

		pnode->child[ pnode->num] = obj;
		pnode->num ++;

		//Adjust the tree.
		AdjustTree( pnode);

		//Update the IF information.
		//adjust_IF( pnode, obj, pnode->num - 1);

/*t/	
		print_IF( pnode);
/*t*/

		return NULL;

	}//pnode->num < SS_Tree_v.
	else
	{
		//No free room for the inserted object, need split node.
		node_t* node_2;
		
		//SplitNode: QuadraticSplit.
		SplitNode( pnode, node_2, obj, SPLIT_OPT);

		if( pnode->level == 0)
			SS_Tree_v.leaf_n ++;
		else
			SS_Tree_v.inner_n ++;

		//Construct the IFs for the newly-created nodes.
		//const_IF( pnode);
		//const_IF( node_2);
/*t/
		int i;
		for( i=0; i<pnode->num; i++)
			print_MBR( pnode->MBR_array[ i], SS_Tree_v.dim);
		for( i=0; i<node_2->num; i++)
			print_MBR( node_2->MBR_array[ i], SS_Tree_v.dim);
/*t*/
/*t/
		print_IF( pnode);
		print_IF( node_2);
/*t*/

		//Insert the split node recursively.
		if( pnode->parent == NULL)
		{	
			//The current node is the root node.
			//Create a new root.
			return CreateNewRoot( pnode, node_2);
		}

		//pnode->parent != NULL
		node_t* pP;
		pP = pnode->parent;
		pP->child[ pnode->loc] = pnode;	//pP->child[ pL->loc] might be invalid due to
										//the "split" operation on pL.

		//UpdateMBR( pnode, pP->MBR_array[ pnode->loc]);
		AdjustTree( pnode);

		//adjust_parent_IF( pP, pnode->loc);

/*t/
		check_and_print_tree( 2);
/*t*/
		
		return InsertSub( pP, node_2);
	}
}

node_t* InsertSub_ss( node_t* pnode, void* obj)
{
	if( pnode->num < M)
	{
		//pnode has room for the new entry.
		if( pnode->level > 0)
		{
			//iner-node case.
			pnode->MBC_array[ pnode->num] = alloc_sphere( SS_Tree_v.dim);
			calc_MBC_ss( ( node_t*)obj, pnode->MBC_array[ pnode->num], 0);
			
			( (node_t*)obj)->parent = pnode;
			( (node_t*)obj)->loc = pnode->num;                      //bug.
		}
		else
		{
			//leaf-node case.
			pnode->MBC_array[ pnode->num] = ( ( obj_t*)obj)->MBC;
		}

		pnode->child[ pnode->num] = obj;
		pnode->num ++;

		//Adjust the tree.
		AdjustTree_ss( pnode);

		//Update the IF information.
		//adjust_IF( pnode, obj, pnode->num - 1);

/*t/	
		print_IF( pnode);
/*t*/

		return NULL;

	}//pnode->num < SS_Tree_v.
	else
	{
		//No free room for the inserted object, need split node.
		node_t* node_2;
		
		//SplitNode: QuadraticSplit.
		SplitNode_ss( pnode, node_2, obj);

		if( pnode->level == 0)
			SS_Tree_v.leaf_n ++;
		else
			SS_Tree_v.inner_n ++;

		//Construct the IFs for the newly-created nodes.
		//const_IF( pnode);
		//const_IF( node_2);
/*t/
		int i;
		for( i=0; i<pnode->num; i++)
			print_MBR( pnode->MBR_array[ i], SS_Tree_v.dim);
		for( i=0; i<node_2->num; i++)
			print_MBR( node_2->MBR_array[ i], SS_Tree_v.dim);
/*t*/
/*t/
		print_IF( pnode);
		print_IF( node_2);
/*t*/

		//Insert the split node recursively.
		if( pnode->parent == NULL)
		{	
			//The current node is the root node.
			//Create a new root.
			return CreateNewRoot_ss( pnode, node_2);
		}

		//pnode->parent != NULL
		node_t* pP;
		pP = pnode->parent;
		pP->child[ pnode->loc] = pnode;	//pP->child[ pL->loc] might be invalid due to
										//the "split" operation on pL.

		//UpdateMBR( pnode, pP->MBR_array[ pnode->loc]);
		AdjustTree_ss( pnode);

		//adjust_parent_IF( pP, pnode->loc);

/*t/
		check_and_print_tree( 2);
/*t*/
		
		return InsertSub_ss( pP, node_2);
	}
}

/*
* AddjustTree is used to adjust the rtree upward after inserting.
*
*  pnode indicates the current node.
* 
* return void.
*/
void AdjustTree( node_t* pnode)
{
	node_t* pP;

	pP = pnode->parent;
	while( pP != NULL)
	{
		if( !UpdateMBR( pnode, pP->MBR_array[ pnode->loc]))
			return;
		else
		{
			pnode = pP;
			pP = pnode->parent;
		}			
	}
}

void AdjustTree_ss( node_t* pnode)
{
	node_t* pP;

	pP = pnode->parent;
	while( pP != NULL)
	{
		if( !UpdateMBC_ss( pnode, pP->MBC_array[ pnode->loc]))
			return;
		else
		{
			pnode = pP;
			pP = pnode->parent;
		}			
	}
}

/*
* Insert is used to insert object into the rtree(root).
*
*  root indicates the rtree to insert in.
*  obj denotes the object to be inserted.
*
* return void.
**/
void Insert( obj_t* obj)
{
	node_t* pnode, *res;

	//SS_Tree_v.obj_n ++;
	ChooseLeaf( SS_Tree_v.root, obj->MBR, pnode);
	res = InsertSub( pnode, obj );

	if( res != NULL)
	{	
		//The root has been changed.
		SS_Tree_v.root = res;
		SS_Tree_v.height ++;
	}

	return;
}

void Insert_ss( obj_t* obj)
{
	node_t* pnode, *res;

	//SS_Tree_v.obj_n ++;
	ChooseLeaf_ss( SS_Tree_v.root, obj->MBC, pnode);
	res = InsertSub_ss( pnode, obj);

	if( res != NULL)
	{	
		//The root has been changed.
		SS_Tree_v.root = res;
		SS_Tree_v.height ++;
	}

	return;
}

/*----------------------Delettion related APIs-------------------------*/

/*
*This function is used to 'Search' the object.
*
*  root indicates the rtree.
*  MBR inidcates the MBR of the object to be searched.
*  loc is used to get back the location of destination in leaf-node.
* 
* return the leaf-node that includes the object, else NULL.
**/
node_t* Search( node_t* root, _range* MBR, int& loc)
{
	int i;
	node_t* node_c;
	
	if( root->level > 0)
	{
		//if the current node is non-leaf node.
		for( i=0; i<root->num; i++)
		{
			if( IsOverlapped(root->MBR_array[i], MBR) == 1)
			{
				node_c = Search( ( node_t*)( root->child[ i]), MBR, loc);
				if( node_c != NULL)
					return node_c;
			}
		}

		//return NULL;
		//All children have been traveled.
	}
	else
	{
		//if the current node is leaf-node.
		for(i=0; i<root->num; i++)
		{
			if( IsSame( root->MBR_array[i], MBR))	//bug--IsOverlapped(root->MBR_array[i], MBR)==1
			{
				loc = i;
				return root;
			}
		}
	}//else	

	return NULL;
}

/*
* ChooseInner is used to choose the inner node for Re-Insertion w.r.t. pnode.
*
*  root is the root of rtree.
*  pnode is the sub to inserted.
*  gets back the inner node's pointer.
*
* return void.
**/
void ChooseInner( node_t* root, node_t* pnode, node_t* &inner_node)
{
	int loc;
	_range* MBR;
	
	MBR = ( _range*)malloc( sizeof( _range));
	memset( MBR, 0, sizeof( _range));

	calc_MBR( pnode, MBR);

	while(1)
	{
		if( root->level == pnode->level+1)
		{
			inner_node = root;
			return;
		}

		NeedLeastArea( root, MBR, loc);
		root = ( node_t*)( root->child[loc]);
	}

	free( MBR);
}

/*
* CondenseSub is used as a sub function in CondenseTree to generate the two chains.
* 
*  pnode indicates the node where deletion occurs.
*  h1 is used to get back the objPointer chain (the last vacant node).
*  h2 is used to get back the nodePointer chain (the last vacant node).
* 
* return void.
**/
void CondenseSub( node_t* pnode, objPointer* h1, nodePointer* h2 )
{
	int i;
	node_t* pParent;
	_range* rTmp;

	if( pnode->parent == NULL)
	{	
		//root node.
		return;
	}
	else if( pnode->num >= m)
	{	
		//Adjust the tree.
		AdjustTree( pnode);	
	}
	else	//pnode->num < m
	{
		//pnode is not a "full" enough.
		//Thus, all children of pnode need to be re-inserted.

		//Delete pnode's entry from its parent node.
		pParent = pnode->parent;
		if( pnode->loc != pParent->num-1)
		{	
			rTmp = pParent->MBR_array[ pnode->loc];
			pParent->child[ pnode->loc] = pParent->child[ pParent->num-1];	//child
			pParent->MBR_array[ pnode->loc] = pParent->MBR_array[ pParent->num-1];		//MBR_array
			
			( ( node_t*)( pParent->child[ pnode->loc]))->loc = pnode->loc;		//
		}
		else
		{
			rTmp = pParent->MBR_array[ pParent->num-1];
		}

		pParent->num --;
		free( rTmp);	
		
		//Add the entries in pnode into two chains.
		if(pnode->level > 0)
		{
			//pnode is a non-leaf node.
			//Add the entries in pnode into head2 (node chain).
			for(i=0; i<pnode->num; i++)
			{
				h2->p = ( node_t*)( pnode->child[ i]);
				h2->next = ( nodePointer*)malloc( sizeof( nodePointer));
				memset( h2->next, 0, sizeof( nodePointer));

				h2 = h2->next;
			}
		}
		else	//pnode->level == 0
		{
			//pnode is a leaf node.
			//Add the entries in pnode into head1 (obj chain).
			for(i=0; i<pnode->num; i++)
			{
				h1->p = ( obj_t*)( pnode->child[i]);
				h1->next = ( objPointer*)malloc( sizeof( objPointer));
				memset( h1->next, 0, sizeof( objPointer));

				h1 = h1->next;
			}
		}
		
		//Release the pnode.
		ReleaseNode( pnode, 1);

		if( pnode->level == 0)
			SS_Tree_v.leaf_n --;
		else
			SS_Tree_v.inner_n --;

		//Condense upward recursively.
		CondenseSub( pParent, h1, h2);
	}//else
}

/*
* CondenseTree is used to condense the tree after deletion operation.
* 
*  pnode indicates the node, where deletion has occured.
*
* return void.
*/ 
void CondenseTree( node_t* pnode)
{
	objPointer* head1, *h1, *eTag1;
	nodePointer* head2, *h2, *eTag2;

	//Create two chains to store the "re-insert entries".
	head1 = ( objPointer*)malloc( sizeof( objPointer));
	memset( head1, 0, sizeof( objPointer));
	head2 = ( nodePointer*)malloc( sizeof( nodePointer));
	memset( head2, 0, sizeof( nodePointer));

	h1 = head1;
	h2 = head2;
	
	//CondenseSub.
	CondenseSub( pnode, h1, h2);
	
	//Perform the re-insertion operations for head1.
	h1 = head1;
	eTag1 = h1->next;
	while( h1 != NULL)
	{
//		if( R_TREE_OR_R_STAR_TREE)	//R-tree
//		{
			node_t* lnode, *res;
			ChooseLeaf( SS_Tree_v.root, h1->p->MBR, lnode);
			res = InsertSub( lnode, h1->p);
			if( res)
			{
				SS_Tree_v.root = res;
				SS_Tree_v.height ++;
			}
//		}
//		else						//R*-tree.
//			RStarInsert( h1->p, 0);		

		free(h1);

		if( eTag1 == NULL)
			break;

		h1 = eTag1;
		eTag1 = h1->next;
	}

	free(h1);

	//Re-Insertion operation for head2.
	h2 = head2;
	eTag2 = h2->next;
	while( h2 != NULL)
	{
//		if( R_TREE_OR_R_STAR_TREE)	//R-tree.
//		{
			node_t* inNode, *res;

			ChooseInner( SS_Tree_v.root, h2->p, inNode);
			res = InsertSub( inNode, h2->p);
			if( res)
			{
				//The root has been changed.
				SS_Tree_v.root = res;
				SS_Tree_v.height ++;
			}
//		}
//		else						//R*-tree.
//			RStarInsert( h2->p, h2->p->level+1);

		free(h2);

		if( eTag2 == NULL)
			break;

		h2 = eTag2;
		eTag2 = h2->next;
	}	

	free(h2);
}

/*
* Delete is used to delete an entry from the rtree.
* 
*  obj donates the object to be deleted.
*
* return 1 if successful; otherwise return 0.
*/
int Delete( obj_t* obj)
{
	int loc;
	node_t* lnode, *root;
	obj_t* oTmp;

	root = SS_Tree_v.root;
	lnode = Search( root, obj->MBR, loc);

	if( lnode == NULL)
	{	
		//The target object does not exist.
		return 0;
	}

	//Delete the object from the node.
	oTmp = ( obj_t*)( lnode->child[loc]);

	lnode->child[ loc] = lnode->child[ lnode->num-1];
	lnode->MBR_array[ loc] = lnode->MBR_array[ lnode->num-1];
	lnode->num --;
	SS_Tree_v.obj_n --;

	free( oTmp->MBR);
	free( oTmp);

	//
	
	//CondenseTree.
	CondenseTree( lnode);
	
	//Update the level if necessary (the root has only one child).
	if( SS_Tree_v.root->num == 1 && SS_Tree_v.root->level > 0)
	{
		lnode = SS_Tree_v.root; 
		SS_Tree_v.root = ( node_t*)( SS_Tree_v.root->child[0]);
		SS_Tree_v.root->parent = NULL;
		SS_Tree_v.root->loc = 0;
		SS_Tree_v.height --;

		//Release the original root node.
		ReleaseNode( lnode, 1);

		SS_Tree_v.inner_n --;
	}
	else if( SS_Tree_v.root->num == 0)
	{	
		//Extreme case.
		SS_Tree_v.height = 0;
	}

	return 1;
}

/*---------------------------Other Operation APIs---------------------------------*/

/*
* ini_tree is used to initialize a rtree (from the datafile, if there exists).
*
* return true if success, else false.
*/
void ini_tree( )
{
	//
	//SS_Tree_v.dim = cfg->dim;
	//M = cfg->M;
	//m = cfg->m;
	//SS_Tree_v.n_key = cfg->key_n;
	
	//strcpy( SS_Tree_v.tree_file, cfg->tree_file);
	//strcpy( SS_Tree_v.data_file, cfg->data_file);

	/*
	if( t_tag == 2)
	{
		SS_Tree_v.height_max = cfg->height_max;
		SS_Tree_v.is_first = ( bool*)malloc( sizeof( bool) * SS_Tree_v.height_max);
		for( i=0; i<SS_Tree_v.height_max; i++)
			SS_Tree_v.is_first[i] = true;
	}
	*/
	
	memset( &SS_Tree_v, 0, sizeof( IRTree_t));

	CreateNode( SS_Tree_v.root);
	SS_Tree_v.leaf_n ++;
}

void ini_tree_ss( )
{
	//
	//SS_Tree_v.dim = cfg->dim;
	//M = cfg->M;
	//m = cfg->m;
	//SS_Tree_v.n_key = cfg->key_n;
	
	//strcpy( SS_Tree_v.tree_file, cfg->tree_file);
	//strcpy( SS_Tree_v.data_file, cfg->data_file);

	/*
	if( t_tag == 2)
	{
		SS_Tree_v.height_max = cfg->height_max;
		SS_Tree_v.is_first = ( bool*)malloc( sizeof( bool) * SS_Tree_v.height_max);
		for( i=0; i<SS_Tree_v.height_max; i++)
			SS_Tree_v.is_first[i] = true;
	}
	*/
	
	memset( &SS_Tree_v, 0, sizeof( IRTree_t));

	CreateNode_ss( SS_Tree_v.root);
	SS_Tree_v.leaf_n ++;
}

/*
* Close is used to *Test* tree, save the rtree ( rTreeSaveFile), and free the resource.
*
*  cfg involves the configuration info.
*  print the tree to the file if o_tag = 1;
* otherwise, print the tree in the console.
*
* return true if succeed; otherwise return false.
*/
bool check_and_print_tree( int o_tag)
{
	int i, j, cnt = 1, inner_cnt, leaf_cnt;
	float min, max;
	FILE* fp;
	
	//Using the 'Circle Queue' to enlarge the space utility.
	if( o_tag == 1)
		fp = fopen( R_TREE_FILE, "w");
	else 
		fp = stdout;
	if( fp == NULL)
	{
		fprintf( stderr, "Cannot create Save_File.\n");
		exit(0);
	}
	
	fprintf( fp, "                                                  R-Tree Information\n");
	fprintf( fp, "M\tm\tInner-nodes\tLeaf-nodes\tObjects\tDimension\n");
	fprintf( fp, "%i\t%i\t%i\t\t\t%i\t\t%i\t%i\n\n", 
				M, m, SS_Tree_v.inner_n, SS_Tree_v.leaf_n, SS_Tree_v.obj_n, SS_Tree_v.dim);
	fprintf( fp, "num   level   loc addr              parent\n");
	
	int sta, rear;
	node *c_node, *root;
	node_t** queue;

	queue = ( node_t**)malloc( ( SS_Tree_v.obj_n + 1) * sizeof( node_t*));
	memset( queue, 0, ( SS_Tree_v.obj_n + 1) * sizeof( node_t*));
	
	sta = rear = 0;
	root = SS_Tree_v.root;
	queue[ rear] = root;
	rear = ( rear + 1) % ( SS_Tree_v.obj_n + 1);	//macro's pit.

	inner_cnt = leaf_cnt = 0;
	while( sta != rear)
	{
		c_node = queue[ sta];
		sta = ( sta + 1) % ( SS_Tree_v.obj_n + 1);

		if( c_node->level == 0)
			leaf_cnt ++;
		else
			inner_cnt ++;
		
		//Print the information of c_node.
		fprintf( fp, "%-4i  %-4i    %-4i%-12X      %-8X\n", 
			c_node->num, c_node->level, c_node->loc, c_node, c_node->parent);
		
		if( c_node->num == 0)
			break;
		for( i=0; i<c_node->num; i++)
		{
			if( c_node->level > 0)
				fprintf( fp, "%-18X", c_node->child[i]);
			else
				fprintf( fp, "%-18i", ( ( obj_t*)(c_node->child[ i]))->id);
		}

		fprintf( fp, "%-8s  %-8s", "min", "max");
		fprintf( fp, "\n");
		
		//Print the 'child' and 'MBR' info.
		for( j=0; j<SS_Tree_v.dim; j++)
		{	
			min = FLT_MAX;
			max = -FLT_MAX;

			for( i=0; i<c_node->num; i++)
			{
				fprintf( fp, "%-8.2f  %-8.2f", 
					c_node->MBR_array[i][j].min, c_node->MBR_array[i][j].max );

				if( c_node->MBR_array[i][j].min < min)
					min = c_node->MBR_array[i][j].min;
				if( c_node->MBR_array[i][j].max > max)
					max = c_node->MBR_array[i][j].max;
			}
/*t*/
			//Test the MBR informatin of the tree.
			if( c_node->parent != NULL)
			{
				if( !( min == c_node->parent->MBR_array[ c_node->loc][j].min && 
					max == c_node->parent->MBR_array[ c_node->loc][j].max))
					fprintf( stderr, "**");
			}
/*t*/
			fprintf( fp, "%-8.2f  %-8.2f", min, max);
			fprintf( fp, "\n");
		}

		//Print the IF info.
		/*
		if( o_tag == 1)
			print_IF( c_node, fp, 2);
		else
			print_IF( c_node, stdout, 2);
		*/
		
		//Enqueue the children of the current node.
		if( c_node->level > 0)
		{
			for(i=0; i<c_node->num; i++)
			{
				queue[ rear] = ( node_t*)( c_node->child[i]);
				rear = ( rear + 1) % ( SS_Tree_v.obj_n + 1);	//macro's pit.

				//Test the 'loc' and 'level' information.
/*t*/
				if( ((node_t*)(c_node->child[i]))->loc != i)
					fprintf( stderr, "The %i'th child's loc is wrong.\n", i+1);
				if( ((node_t*)(c_node->child[i]))->level != c_node->level-1)
					fprintf( stderr, "The %i'th child's level is wrong.\n", i+1);
/*t*/
			}
		}

		fprintf( fp, "\n");
	}//while
	
	free( queue);
	if( o_tag == 1)
		fclose( fp);

	if( SS_Tree_v.inner_n != inner_cnt)
		printf( "Inconsistent inner_n info!\n");
	if( SS_Tree_v.leaf_n != leaf_cnt)
		printf( "Inconsistent leaf_n info!\n");

	return true;
}

bool check_and_print_tree_ss( int o_tag)
{
	int i, j, cnt = 1, inner_cnt, leaf_cnt;
	//float min, max;
	FILE* fp;
	
	//Using the 'Circle Queue' to enlarge the space utility.
	if( o_tag == 1)
		fp = fopen( SS_TREE_FILE, "w");
	else 
		fp = stdout;
	if( fp == NULL)
	{
		fprintf( stderr, "Cannot create %s.\n", SS_TREE_FILE);
		exit(0);
	}
	
	fprintf( fp, "                                                  SS-Tree Information\n");
	fprintf( fp, "M\tm\tInner-nodes\tLeaf-nodes\tObjects\tDimension\n");
	fprintf( fp, "%i\t%i\t%i\t\t\t%i\t\t%i\t%i\n\n", 
				M, m, SS_Tree_v.inner_n, SS_Tree_v.leaf_n, SS_Tree_v.obj_n, SS_Tree_v.dim);
	fprintf( fp, "num   level   loc addr              parent\n");
	
	int sta, rear;
	node *c_node, *root;
	node_t** queue;

	queue = ( node_t**)malloc( ( SS_Tree_v.obj_n + 1) * sizeof( node_t*));
	memset( queue, 0, ( SS_Tree_v.obj_n + 1) * sizeof( node_t*));
	
	sta = rear = 0;
	root = SS_Tree_v.root;
	queue[ rear] = root;
	rear = ( rear + 1) % ( SS_Tree_v.obj_n + 1);	//macro's pit.

	inner_cnt = leaf_cnt = 0;
	while( sta != rear)
	{
		c_node = queue[ sta];
		sta = ( sta + 1) % ( SS_Tree_v.obj_n + 1);

		if( c_node->level == 0)
			leaf_cnt ++;
		else
			inner_cnt ++;
		
		//Print the information of c_node.
		fprintf( fp, "%-4i\t%-4i\t%-4i%-12X\t%-8X\n", 
			c_node->num, c_node->level, c_node->loc, c_node, c_node->parent);
		
		if( c_node->num == 0)
			break;
		for( i=0; i<c_node->num; i++)
		{
			if( c_node->level > 0)
				fprintf( fp, "%-18X", c_node->child[i]);
			else
				fprintf( fp, "%-18i", ( ( obj_t*)(c_node->child[ i]))->id);
		}

		//fprintf( fp, "%-8s  %-8s", "min", "max");
		fprintf( fp, "\n");
		
		//Print the 'child' and 'MBR' info.
		for( j=0; j<SS_Tree_v.dim; j++)
		{	
			//min = FLT_MAX;
			//max = -FLT_MAX;

			for( i=0; i<c_node->num; i++)
			{
				//fprintf( fp, "%-8.2f  %-8.2f", 
				//	c_node->MBR_array[i][j].min, c_node->MBR_array[i][j].max );

				fprintf( fp, "%-8.2f\t%-8.2f\t", 
					c_node->MBC_array[ i]->center->coord[ j] - c_node->MBC_array[ i]->radius, 
					c_node->MBC_array[ i]->center->coord[ j] + c_node->MBC_array[ i]->radius);

				//if( c_node->MBR_array[i][j].min < min)
				//	min = c_node->MBR_array[i][j].min;
				//if( c_node->MBR_array[i][j].max > max)
				//	max = c_node->MBR_array[i][j].max;
			}
/*t*/
			//Test the MBR informatin of the tree.
			if( c_node->parent != NULL)
			{
				//if( !( min == c_node->parent->MBR_array[ c_node->loc][j].min && 
				//	max == c_node->parent->MBR_array[ c_node->loc][j].max))
				//	fprintf( stderr, "**");

				//
				if( UpdateMBC_ss( c_node, c_node->parent->MBC_array[ c_node->loc]))
					fprintf( stderr, "**");
			}
/*t*/
			//fprintf( fp, "%-8.2f  %-8.2f", min, max);
			fprintf( fp, "\n");
		}

		//Print the IF info.
		/*
		if( o_tag == 1)
			print_IF( c_node, fp, 2);
		else
			print_IF( c_node, stdout, 2);
		*/
		
		//Enqueue the children of the current node.
		if( c_node->level > 0)
		{
			for(i=0; i<c_node->num; i++)
			{
				queue[ rear] = ( node_t*)( c_node->child[i]);
				rear = ( rear + 1) % ( SS_Tree_v.obj_n + 1);	//macro's pit.

				//Test the 'loc' and 'level' information.
/*t*/
				if( ((node_t*)(c_node->child[i]))->loc != i)
					fprintf( stderr, "The %i'th child's loc is wrong.\n", i+1);
				if( ((node_t*)(c_node->child[i]))->level != c_node->level-1)
					fprintf( stderr, "The %i'th child's level is wrong.\n", i+1);
/*t*/
			}
		}

		fprintf( fp, "\n");
	}//while
	
	free( queue);
	if( o_tag == 1)
		fclose( fp);

	if( SS_Tree_v.inner_n != inner_cnt)
		printf( "Inconsistent inner_n info!\n");
	if( SS_Tree_v.leaf_n != leaf_cnt)
		printf( "Inconsistent leaf_n info!\n");

	return true;
}

/*
 *	Un-set attributes: child and parent.
 */
void read_node( FILE* i_fp, node_t* node_v)
{
	int i, j;
	//B_KEY_TYPE key;
	float tmp_1, tmp_2;
	//char tmp_c;
	char des[ MAX_LINE_LENG];
	//bst_node_t* bst_node_v;

	//CreateNode( node_v);

	/*Only applied to WIN32.
	//fgets( des, MAX_LINE_LENG, i_fp);	//blank line.
	//fgetc( i_fp);
	fgets( des, MAX_LINE_LENG, i_fp);
	//fscanf( i_fp, "\n");
	fscanf( i_fp, "%i%i%i", &node_v->num, &node_v->level, &node_v->loc);
	fgets( des, MAX_LINE_LENG, i_fp);
	fgets( des, MAX_LINE_LENG, i_fp);
	*/

	fscanf( i_fp, "%i%i%i", &node_v->num, &node_v->level, &node_v->loc);

	/*t/
	printf( "%i\t%i\t%i\t", node_v->num, node_v->level, node_v->loc);
	/*t*/

	for( i=0; i<2+node_v->num+2; i++)
	{
		fscanf( i_fp, "%s", des);
		/*t/
		printf( "%s\t", des);
		/*t*/
	}
	//printf( "\n");
	
	//Read the MBR info.
	for( i=0; i<node_v->num; i++)
	{
		node_v->MBR_array[ i] = ( _range*)malloc( sizeof( _range));
		memset( node_v->MBR_array[ i], 0, sizeof( _range));
	}
	
	for( j=0; j<SS_Tree_v.dim; j++)
	{
		for( i=0; i<node_v->num; i++)
		{
			fscanf( i_fp, "%f%f", &node_v->MBR_array[ i][j].min, &node_v->MBR_array[ i][j].max);
			/*t/
			printf( "%f\t%f\t", node_v->MBR_array[ i][j].min, node_v->MBR_array[ i][j].max);
			/*t*/
		}
		fscanf( i_fp, "%f%f", &tmp_1, &tmp_2);
		/*t/
		printf( "%f\t%f\n", tmp_1, tmp_2);
		/*t*/
	}
}

/*
 *
 */
obj_t* const_obj( node_t* node_v, int loc, int& id_cnt)
{
	//int tag;
	//bst_node_t* x;
	obj_t* obj_v;
//	k_node_t* k_node_v;

	obj_v = ( obj_t*)malloc( sizeof( obj_t));
	memset( obj_v, 0, sizeof( obj_t));
	
	//id & MBr attributes.
	obj_v->id = id_cnt++;
	obj_v->MBR = node_v->MBR_array[ loc];

	//Initialize the k_head attribute.
	/*
	obj_v->k_head = ( k_node_t*)malloc( sizeof( k_node_t));
	memset( obj_v->k_head, 0, sizeof( k_node_t));
	k_node_v = obj_v->k_head;

	tag = 0;
	x = node_v->bst_v->root;
	
	while( get_next_in_order( x, tag))
	{
		//For a specific IF entry (a specific keyword x->key).		
		if( get_k_bit( x->p_list, loc))
			add_keyword_entry( k_node_v, x->key);
		
		if( !in_order_sub( x, tag))
			break;
	}
	*/

	return obj_v;
}


/*
 *	Read the IR-tree from the file SAVE_FILE.
 *	The required file is the output file of "check_and_print_tree".
 */
void read_tree( char* ir_tree_f)
{
	int i, j, inner_node_n, leaf_node_n, obj_n, dim, sta, rear, id_cnt, tmp;
	char des[ MAX_LINE_LENG];
	FILE* i_fp;
	node_t** queue;
	node_t* c_node, *n_node;
	obj_t* obj_v;

	if( ( i_fp = fopen( ir_tree_f, "r")) == NULL)
	{
		printf( "The ir_tree_f cannot open.\n");
		exit( 0);
	}

	/*Only applied to WIN32.
	fgets( des, MAX_LINE_LENG, i_fp);				//line 1.
	fgets( des, MAX_LINE_LENG, i_fp);				//line 2.
	fscanf( i_fp, "%i%i%i%i%i%i", &tmp, &tmp, &inner_node_n, &leaf_node_n, &obj_n, &dim);	//line 3.
	fgets( des, MAX_LINE_LENG, i_fp);				//line 4.
	fgets( des, MAX_LINE_LENG, i_fp);				//line 5.
	*/

	for( i=0; i<8; i++)
	{
		fscanf( i_fp, "%s", des);
		/*t/
		printf( "%s\t", des);
		/*t*/
	}
	//printf( "\n");

	fscanf( i_fp, "%i%i%i%i%i%i", &tmp, &tmp, &inner_node_n, &leaf_node_n, &obj_n, &dim);	//line 3.

	/*t/
	printf( "%i\t%i\t%i\t%i\t%i\t%i\t", tmp, tmp, inner_node_n, leaf_node_n, obj_n, dim);	
	/*t*/

	for( i=0; i<5; i++)
	{
		fscanf( i_fp, "%s", des);
		/*t/
		printf( "%s\t", des);
		/*t*/
	}
	//printf( "\n");

	queue = ( node_t**)malloc( inner_node_n * sizeof( node_t*));
	memset( queue, 0, inner_node_n * sizeof( node_t*));

	sta = 0;
	rear = -1;
	id_cnt = 1;

	//Initialize the IR-tree.
	ini_tree( );
	SS_Tree_v.dim = dim;

	SS_Tree_v.leaf_n = 0;	//bug.

	/*
	SS_Tree_v.inner_n = inner_node_n;
	SS_Tree_v.leaf_n = leaf_node_n;
	SS_Tree_v.obj_n = obj_n;
	*/

	//Read the root node.
	read_node( i_fp, SS_Tree_v.root);
	SS_Tree_v.root->parent = NULL;
	if( SS_Tree_v.root->level > 0)
	{
		//
		queue[ ++rear] = SS_Tree_v.root;
		SS_Tree_v.inner_n ++;
	}
	else
	{
		//The root is a leaf node.
		//Create the corresponding the objects.
		for( i=0; i<SS_Tree_v.root->num; i++)
		{
			obj_v = const_obj( SS_Tree_v.root, i, id_cnt);
			
			SS_Tree_v.root->child[ i] = obj_v;
		}			

		SS_Tree_v.leaf_n ++;
		SS_Tree_v.obj_n += SS_Tree_v.root->num;
	}

	while( sta <= rear)
	{
		c_node = queue[ sta++];
		
		/*t/
		printf( "sta: %i\n", sta);
		/*t*/
		/*t/
		if( sta == 5)
			printf( "");
		/*t*/

		//Read the node info.
		for( i=0; i<c_node->num; i++)
		{
			CreateNode( n_node);

			//fgets( des, MAX_LINE_LENG, i_fp);

			read_node( i_fp, n_node);
			
			n_node->parent = c_node;
			c_node->child[ i] = n_node;

			//Put n_node into the queue if n_node is an inner-node.
			if( n_node->level > 0)
			{
				queue[ ++rear] = n_node;

				SS_Tree_v.inner_n ++;
			}
			else
			{
				//n_node is a leaf-node.
				//Create the objects.
				for( j=0; j<n_node->num; j++)
				{
					obj_v = const_obj( n_node, j, id_cnt);

					n_node->child[ j] = obj_v;
				}

				SS_Tree_v.leaf_n ++;
				SS_Tree_v.obj_n += n_node->num;

				/*t/
				if( SS_Tree_v.leaf_n == 1113)
					printf( "");
				/*t*/
			}
		}//for
	}//while

	if( SS_Tree_v.leaf_n != leaf_node_n)
		printf( "Leaf node number inconsistency..\n");
	if( SS_Tree_v.inner_n != inner_node_n)
		printf( "Inner node number inconsistency..\n");
	if( SS_Tree_v.obj_n != obj_n)
		printf( "Object number inconsistency..\n");

	free( queue);
	fclose( i_fp);
}

/*---------------------Added APIs-2011-03-09------------------------*/


/*
*	build_IRTree constructs the IR-tree based on the data file.
*
void build_IRTree( data_t* data_v)
{
	int i;

	//Initialize a IR-tree.
	ini_tree( );
	SS_Tree_v.dim = data_v->dim;

	//Insert all the objects to construct the IR-tree.
	for( i=0; i<data_v->obj_n; i++)
	{
//		if( t_tag == 1)		//R-tree.
//		{
			Insert( data_v->obj_v + i);
//		}
//		else				//R*-tree.
//		{
//			RStarInsert( data_v->obj_v + i, 0);
//		}

		SS_Tree_v.obj_n ++;


//		check_and_print_tree( 2);
		check_IF( );
//
	}
}
*/

/*
 *	Release the sub-tree rooted at node_v.
 */
void free_IRTree_sub( node_t* node_v)
{
	int i;

	if( node_v == NULL)
		return;
	
	if( node_v->level > 0)
	{
		for( i=0; i<node_v->num; i++)
			free_IRTree_sub( ( node_t*)( node_v->child[ i]));
	}

	ReleaseNode( node_v, 1);
}

/*
 *	Re-implementation of free_IRTree.
 *
 *	Method: DFS.
 */
void free_IRTree( )
{
	free_IRTree_sub( SS_Tree_v.root);
}

/*
 *	Print a MBR.
 */
void print_MBR( _range* MBR, int dim)
{
	int i;

	for( i=0; i<dim; i++)
	{
		printf( "%.2f\t%.2f\n", MBR[ i].min, MBR[ i].max);
	}
	printf( "\n");
}

/*
*	test_IRTree verifies newly-added functions related IR-tree.
*	
void test_IRTree( int o_tag)
{
	IRTree_config_t* cfg;
	data_t* data_v;

	//Read the configuration info.
	cfg = read_config_irtree( );

	//Read the data.
	data_v = read_data_irtree( cfg);

	//Build the IR-tree.
	printf( "Building the IR-tree ...\n");
	build_IRTree( data_v);

	//Check the IR-tree.
	//check_IF( );

	//Saving the IR-tree.
	printf( "Saving the IR-tree ...\n");
	check_and_print_tree( o_tag, cfg->tree_file);

	//Release the IR-tree.
	printf( "Releasing the IR-tree ...\n");
	free_IRTree( );
	
	//Release the related the resources.
	free( cfg);
	release_data( data_v);
}
*/


//Supporting APIs for SS-tree;
void build_SSTree( sphere_data_t* sphere_data_v)
{
	int i;
	obj_t* obj_v;

	//Initialize a IR-tree.
	ini_tree_ss( );

	SS_Tree_v.dim = sphere_data_v->dim;

	//Insert all the objects to construct the IR-tree.
	for( i=0; i<sphere_data_v->sphere_n; i++)
	{
//		if( t_tag == 1)		//R-tree.
//		{
			//Insert( data_v->obj_v + i);
//		}
//		else				//R*-tree.
//		{
//			RStarInsert( data_v->obj_v + i, 0);
//		}

		obj_v = create_obj_ss( sphere_data_v->sphere_v[ i], i+1);

		Insert_ss( obj_v);

		SS_Tree_v.obj_n ++;

//		check_and_print_tree( 2);
//		check_IF( );
//
	}
}

obj_t* create_obj_ss( sphere_t* MBC, int id)
{
	obj_t* obj_v;

	obj_v = ( obj_t*)malloc( sizeof( obj_t));
	memset( obj_v, 0, sizeof( obj_t));

	obj_v->MBC = MBC;
	obj_v->id = id;

	return obj_v;
}

void free_SSTree_sub( node_t* node_v)
{
	int i;

	if( node_v == NULL)
		return;
	
	if( node_v->level > 0)
	{
		for( i=0; i<node_v->num; i++)
			free_SSTree_sub( ( node_t*)( node_v->child[ i]));
	}
	else
	{
		for( i=0; i<node_v->num; i++)
			free( ( obj_t*)( node_v->child[ i]));
	}

	ReleaseNode_ss( node_v, 1);
}

void free_SSTree( )
{
	free_SSTree_sub( SS_Tree_v.root);
}

void test_SSTree( )
{
	config_t* config_v;
	sphere_data_t* sphere_data_v;

	//Read the hyperspheres;
	config_v = read_config( );

	sphere_data_v = read_sphere_data( config_v);

	//Construct the SS-Tree;
	build_SSTree( sphere_data_v);

	//Check and Print the tree;
	check_and_print_tree_ss( 1);

	//Release the resources;
	free_SSTree( );
	free( config_v);
	release_sphere_data( sphere_data_v);
}


void traverse_SSTree( node_t* node_v, sphere_list_t* sphere_list_v)
{
	int i;

	if( node_v->level > 0)
	{
		//Recursively traverse the sub-tree located at each child of node_v;
		for( i=0; i<node_v->num; i++)
			traverse_SSTree( ( ( node_t*)node_v->child[ i]), sphere_list_v);
	}
	else
	{
		for( i=0; i<node_v->num; i++)
			add_entry_sphere_list( sphere_list_v, node_v->MBC_array[ i]);
	}
}