
#ifndef SSTREE_H
#define SSTREE_H

#include <limits.h>
#include <stdio.h>
#include <stdlib.h>
#include <math.h>
#include <string.h>
#include <float.h>


//#include "bit_operator.h"
//#include "bst.h"
#include "data_struct.h"
//#include "data_struct3.h"
#include "data_utility.h"

//#define IRTree_OR_R_STAR_TREE	1

#define M				32
#define m				16
#define SPLIT_OPT		1

#define MAX_LINE_LENG	1027

//#define DATA_FILE		"IRTree_DataFile.txt"		//
#define SAVE_FILE		"IRTree.txt"				//savefile of rtree.
#define R_TREE_FILE		"r_tree.txt"
#define SS_TREE_FILE	"ss_tree.txt"


#define max(a, b) ( ( a) > (b) ? ( a) : (b))
#define min(a, b) ( ( a) < (b) ? ( a) : (b))

//The interval structure of one dimension.
typedef struct _range
{
	float min;
	float max;

}	_range_t;

//The structure of the object.
typedef struct obj
{
	int			id;
	
	//For R-tree
	_range*		MBR;

	//For SS-tree
	sphere_t*	MBC;

	//IR-tree augmentation.
	//k_node_t*	k_head;
	
}	obj_t;

//The node structure for storing an object pointer.
typedef struct obj_node
{
	obj_t*				obj_v;
	struct obj_node*	next;

	//for const_k_NN_key only.
	//B_KEY_TYPE			dist;
}	obj_node_t;

//The node structure of the node of the rtree.
typedef struct node         
{
	int			num;            //The number of items in the node.
	int			loc;            //The location in its parent's node.

	int			level;			//from 'leaf(0)' to 'root', increased by 1 level by level.
	node*		parent;			//pointer to the parent node.
	void**		child;			//pointer array (because the inner-node and leaf-node has different chidren) 
	
	//For R-tree
	_range**	MBR_array;		//closing MBRangle within 'x' dimension.

	//For SS-tree
	sphere_t**	MBC_array;		//

	//sphere_t*	MBC;			//"Minimum Bounding Circle" for the current node.
								//Maintained by its parent node (except the root node)

	//Extension for IR-tree.
	//BIT_TYPE *keys;
		//bit array is adopted for representing the posting list.
		//adjacent list is another data structure option.
	//k_node_t** key;

	//bst_t*	bst_v;
	
}	node_t;

//The info structure of the rtree.
typedef struct	IRTree
{
	int		dim;
	
	//R-tree info.
	//int	R_M;		
	//int	R_m;

	int		split_opt;
	
	node_t*	root;			//root node.
	int		height;			//height.
	int		obj_n;			//the number of the objects indexed by the tree.

	int		inner_n;		//the number of inner-nodes.
	int		leaf_n;			//the number of leaf-nodes.

	//bool*	is_first;		//used for the r*-tree.
	//int	height_max;		//used for the r*-tree.

	//IF info.
	//int		key_n;
	
}	IRTree_t;

//
struct objPointer	//typedef 
{
	obj_t*		p;
	objPointer* next;
};

//
struct nodePointer	//typedef 
{
	node_t*			p;
	nodePointer*	next;
};


//Added on 4 Dec. 2013 by Cheng
typedef struct _stat
{
	float memory_v;
	float tree_memory_v;
	float tree_memory_max;

}	_stat_t;

extern IRTree_t SS_Tree_v;
extern _stat_t _stat_v;

/*---------------------Calculation Related APIs----------------------*/


int compare (const void * a, const void * b);

//double abs( double var);

//--
void CreateNode( node_t* &pnode);
void CreateNode_ss( node_t* &pnode);

//--
void ReleaseNode( node_t* pnode, int tag);
void ReleaseNode_ss( node_t* pnode, int tag);

int IsOverlapped( _range* MBR1, _range* MBR2);

bool IsSame( _range* MBR1, _range* MBR2);

double GetArea( node_t* pnode);	
double GetArea_2( _range* MBR1, _range* MBR2);
double GetIncArea( node_t* pnode, _range* MBR);

//--
void NeedLeastArea( node_t* pnode, _range* obj, int& loc);
void NeedLeastArea_ss( node_t* pnode, sphere_t* obj, int& loc);

//--
void calc_MBR( node_t* pnode, _range* &MBR);
void calc_MBC_ss( node_t* pnode, sphere_t* &MBC, int opt);

_range* get_MBR_node( node_t* pnode, int dim);

//--
bool UpdateMBR( node_t* pnode, _range* &MBR);
bool UpdateMBC_ss( node_t* pnode, sphere_t* &MBC);
 
/*----------------------Insertion related APIs----------------------*/

int find_closest_branch( node_t* node_v, sphere_t* MBC);

//--
void ChooseLeaf( node_t* root, _range* MBR, node_t* &leaf_node);
void ChooseLeaf_ss( node_t* root, sphere_t* MBC, node_t* &leaf_node);

//--
void PickSeeds( node_t* pnode, _range* MBR, int& g1, int& g2);
//void PickSeeds_ss( node_t* pnode, sphere_t* MBC, int& g1, int& g2);
void LinearPickSeeds( node_t* pnode, _range* MBR, int& g1, int& g2);
int PickNext( node_t* pnode, node_t* node_1, node_t* node_2, node_t* &node_c, _range* MBR, bool* done, int rTag);
int LinearPickNext( node_t* pnode, node_t* node_1, node_t* node_2, node_t* &node_c, bool* done, int rTag);

//--
void AssignEntry( node_t* pnode, node_t* node_c, void* obj, _range* MBR);
void AssignEntry_ss( node_t* pnode, node_t* node_c, void* obj, sphere_t* MBC);

//--
void SplitNode( node_t* &pnode, node_t* &node_2, void* obj, int opt);
void SplitNode_ss( node_t* &pnode, node_t* &node_2, void* obj);

double calc_variance_ss( COORD_TYPE* values, int size);
COORD_TYPE calc_median_ss( COORD_TYPE* values, int size);
//COORD_TYPE calc_median_ss( node_t* pnode, sphere_t* MBC, int dim);
int* calc_group_indicator_ss( node_t* pnode, sphere_t* MBC, int dim);
int pick_axis_variance_ss( node_t* pnode, sphere_t* MBC);

//void QuadraticSplit( node_t* &pnode, node_t* &node_2, void* obj);
//void LinearSplit( node_t* &pnode, node_t* &node_2, void* obj);

//--
void AdjustTree( node_t* pnode);
void AdjustTree_ss( node_t* pnode);

//--
node_t* CreateNewRoot( node_t* node_1, node_t* node_2);
node_t* CreateNewRoot_ss( node_t* node_1, node_t* node_2);

//--
node_t* InsertSub( node_t* pnode, void* obj);

//--
void Insert( obj_t* obj);
void Insert_ss( obj_t* obj);

/*---------------------Deletion Related APIs------------------------*/

node_t* Search( node_t* root, _range* MBR, int& loc);

void ChooseInner( node_t* root, _range* MBR, node_t* &inner_node);
void CondenseSub( node_t* pnode, objPointer* h1, nodePointer* h2 );
void CondenseTree( node_t* pnode);
int Delete( obj_t* obj);

/*---------------------Other Operation APIs------------------------*/

//--
void ini_tree( );
void ini_tree_ss( );
obj_t* create_obj_ss( sphere_t* MBC, int id);

//--
bool print_and_check_tree( int o_tag);
bool print_and_check_tree_ss( int o_tag);

void read_node( FILE* i_fp, node_t* node_v);

//--
obj_t* const_obj( node_t* node_v, int loc, int& id_cnt);
void read_tree( char* ir_tree_f);

//To be implemented.
//void build_IRTree( data_t* data_v);
void free_IRTree_sub( );
void free_IRTree( );
//void bp_free_IRTree( );
void print_MBR( _range* MBR, int dim);

//void test_IRTree( int o_tag);

//Data interfaces for the SS-tree;
void build_SSTree( sphere_data_t* sphere_data_v);
void free_SSTree_sub( );
void free_SSTree( );
void test_SSTree( );


void traverse_SSTree( node_t* node_v, sphere_list_t* sphere_list_v);

#endif